from __future__ import annotations


__copyright__ = "Copyright (C) 2022 University of Illinois Board of Trustees"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import ast
from collections.abc import Callable, Iterable, MutableMapping
from functools import cached_property, lru_cache
from typing import TextIO, TypeVar, cast


# This machinery applies AST rewriting to the mapper in a mildly brutal
# manner, and as such it requires some attention from the user to
# make sure all transformations applied are valid. A good way to do
# this is to look at the generated code by setting print_modified_code_file
# to sys.stdout or the like.

# Note that this machinery is intentionally generic enough so as to also
# apply to pytato's mappers in unmodified form.


# {{{ ast retrieval

AstDefNodeT = TypeVar("AstDefNodeT", ast.FunctionDef, ast.ClassDef)


def _get_def_from_ast_container(
            container: Iterable[ast.AST],
            name: str,
            node_type: type[AstDefNodeT]
        ) -> AstDefNodeT:
    for entry in container:
        if isinstance(entry, node_type) and entry.name == name:
            return entry

    raise ValueError(f"symbol '{name}' not found")


@lru_cache
def _get_ast_for_file(filename: str) -> ast.Module:
    with open(filename) as inf:
        return ast.parse(inf.read(), filename)


def _get_file_name_for_module_name(module_name: str) -> str | None:
    from importlib import import_module
    return import_module(module_name).__file__


def _get_ast_for_module_name(module_name: str) -> ast.Module:
    return _get_ast_for_file(_get_file_name_for_module_name(module_name))


def _get_module_ast_for_object(obj):
    return _get_ast_for_module_name(obj.__module__)


def _get_ast_for_class(cls: type) -> ast.ClassDef:
    mod_ast = _get_module_ast_for_object(cls)
    return _get_def_from_ast_container(
            mod_ast.body, cls.__name__, ast.ClassDef)


def _get_ast_for_method(f: Callable) -> ast.FunctionDef:
    dot_components = f.__qualname__.split(".")
    assert dot_components[-1] == f.__name__
    cls_name, = dot_components[:-1]
    mod_ast = _get_ast_for_module_name(f.__module__)
    cls_ast = _get_def_from_ast_container(
            mod_ast.body, cls_name, ast.ClassDef)
    return _get_def_from_ast_container(
            cls_ast.body, f.__name__, ast.FunctionDef)

# }}}


def _replace(obj, **kwargs):
    try:
        kwargs["lineno"] = obj.lineno
    except AttributeError:
        pass
    try:
        kwargs["col_offsets"] = obj.col_offsets
    except AttributeError:
        pass

    return type(obj)(**{
        name: kwargs.get(name, getattr(obj, name))
        for name in set(obj._fields) | set(kwargs.keys())
        })


class _VarArgsRemover(ast.NodeTransformer):
    def __init__(self, drop_args, drop_kwargs):
        self.drop_args = drop_args
        self.drop_kwargs = drop_kwargs

    def visit_Call(self, node):  # noqa: N802
        node = self.generic_visit(node)
        return _replace(node,
                       args=[arg for arg in node.args
                          if not self.drop_args or not isinstance(arg, ast.Starred)],
                       keywords=[kw for kw in node.keywords
                          if not self.drop_kwargs or kw.arg is not None])


class _RecInliner(ast.NodeTransformer):
    def __init__(self, *, inline_rec, inline_cache):
        self.inline_rec = inline_rec
        self.inline_cache = inline_cache

    def visit_Call(self, node: ast.Call) -> ast.AST:  # noqa: N802
        node = cast(ast.Call, self.generic_visit(node))

        result_expr: ast.expr = node

        if (isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "self"
                and node.func.attr in ["rec", "rec_arith"]):

            from ast import (
                Attribute,
                Call,
                Compare,
                Constant,
                IfExp,
                IsNot,
                Load,
                Name,
                NamedExpr,
                Store,
            )
            expr = node.args[0]
            self_sym = Name(id="self", ctx=Load())
            fallback_call = _replace(
                    node,
                    func=Attribute(value=self_sym, attr="rec_fallback", ctx=Load()))

            def getattr_sym(obj, name):
                return Call(
                        func=Name(id="getattr", ctx=Load()),
                        args=[obj, name, Constant(value=None)], keywords=[])

            def is_not_none(obj):
                return Compare(
                        left=obj, ops=[IsNot()], comparators=[Constant(value=None)])

            def expr_assign(name, value):
                return NamedExpr(
                        target=Name(id=name, ctx=Store()),
                        value=value)

            if self.inline_rec:
                result_expr = IfExp(
                    test=is_not_none(
                        expr_assign("mname",
                            getattr_sym(expr, Constant(value="mapper_method")))),
                    body=IfExp(
                        test=is_not_none(
                            expr_assign("method", getattr_sym(
                                    self_sym, Name(id="mname", ctx=Load())))),
                        body=_replace(node, func=Name(id="method", ctx=Load())),
                        orelse=fallback_call),
                    orelse=fallback_call)

            if self.inline_cache:
                cache = Attribute(value=self_sym, attr="_cache")
                expr_type = Call(
                        func=Name(id="type", ctx=Load()),
                        args=[expr],
                        keywords=[])
                cache_key_expr = ast.Tuple([expr_type, expr], ctx=Load())
                nic = Name(id="_NotInCache", ctx=Load())

                result_expr = IfExp(
                        test=Compare(
                            left=expr_assign(
                                "result",
                                Call(
                                    func=Attribute(value=cache, attr="get"),
                                    args=[
                                        expr_assign(
                                            "cache_key",
                                            cache_key_expr),
                                        nic
                                        ], keywords=[])),
                                ops=[IsNot()], comparators=[nic]),
                        body=Name(id="result", ctx=Load()),
                        orelse=Call(
                            func=Name(id="_set_and_return", ctx=Load()),
                            args=[cache,
                                    Name(id="cache_key", ctx=Load()),
                                    result_expr], keywords=[]))

        return result_expr


def _get_cache_key_expr(mdef):
    for stmt in mdef.body:
        if isinstance(stmt, ast.Expr):
            # assume no side effects, likely a docstring
            pass
        elif isinstance(stmt, ast.Return):
            return stmt.value
        else:
            raise ValueError("unexpected statement type in get_cache_key: "
                             "{type(stmt).__name__} --- must only contain a single "
                             "return statement")


class _CacheKeyInliner(ast.NodeTransformer):
    def __init__(self, *, cache_key_expr):
        self.cache_key_expr = cache_key_expr

    def visit_Call(self, node):  # noqa: N802
        node = self.generic_visit(node)

        result_expr = node

        if (isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "self"
                and node.func.attr == "get_cache_key"):
            result_expr = self.cache_key_expr

        return result_expr


KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")


def _set_and_return(
            mapping: MutableMapping[KeyT, ValueT],
            key: KeyT,
            value: ValueT
        ) -> ValueT:
    mapping[key] = value
    return value


def optimize_mapper(
        *, drop_args: bool = False, drop_kwargs: bool = False,
        inline_rec: bool = False, inline_cache: bool = False,
        inline_get_cache_key: bool = False,
        print_modified_code_file: TextIO | None = None) -> Callable[[type], type]:
    """
    :param print_modified_code_file: a file-like object to which the modified
        code will be printed, or ``None``.
    """
    # This is a crime, an abomination. But a somewhat effective one.

    def wrapper(cls: type) -> type:
        try:
            # Introduced in Py3.9
            ast.unparse  # noqa: B018
        except AttributeError:
            return cls

        # This also needs Py3.8 for the walrus operators used in inlined rec.

        cls_ast = _get_ast_for_class(cls)

        # {{{ gather relevant method definitions

        method_defs = {}
        other_contents = []

        for entry in cls_ast.body:
            if isinstance(entry, ast.FunctionDef):
                method_defs[entry.name] = entry
            else:
                other_contents.append(entry)

        seen_module_names = set()

        # Yes, this grabs the source for all methods, including those
        # from the base classes, and rewrites them. Otherwise, those methods
        # would still use have *args, **kwargs.
        for name in dir(cls):
            if not name.startswith("__") or name == "__call__":
                method = getattr(cls, name)
                if (not callable(method)
                        or isinstance(method, property | cached_property)):
                    # properties don't have *args, **kwargs
                    continue

                seen_module_names.add(method.__module__)
                method_ast = _get_ast_for_method(method)
                if name != method_ast.name:
                    # This happens for aliases. Make them separate methods.
                    method_ast = _replace(method_ast, name=name)

                method_defs[method_ast.name] = method_ast

        # }}}

        # {{{ get expression for get_cache_key

        cache_key_expr = None
        if inline_get_cache_key and "get_cache_key" in method_defs:
            cache_key_expr = _get_cache_key_expr(method_defs["get_cache_key"])

        if inline_get_cache_key and cache_key_expr is None:
            raise ValueError("could not find expression for cache key")

        # }}}

        # {{{ rewrite method_defs

        new_method_defs = []

        for mname in sorted(method_defs):
            mdef = method_defs[mname]

            mdef = _replace(mdef,
                    args=_replace(mdef.args,
                        vararg=None if drop_args else mdef.args.vararg,
                        kwarg=None if drop_kwargs else mdef.args.kwarg))

            mdef = _VarArgsRemover(
                    drop_args=drop_args, drop_kwargs=drop_kwargs).visit(mdef)

            if cache_key_expr is not None:
                mdef = _CacheKeyInliner(cache_key_expr=cache_key_expr).visit(mdef)

            mdef = _RecInliner(
                    inline_rec=inline_rec, inline_cache=inline_cache).visit(mdef)

            ast.fix_missing_locations(mdef)

            new_method_defs.append(mdef)

        cls_ast = _replace(
                cls_ast,
                # other contents second so method aliases work
                body=new_method_defs + other_contents,
                # FIXME: only remove optimize_mapper from decorators
                decorator_list=[])

        # }}}

        code_str = (
                # Incoming code *may* rely on deferred evaluation of annotations.
                # Since we're not checking whether it does, turn it on just in case.
                "from __future__ import annotations\n"
                + ast.unparse(cls_ast))
        if print_modified_code_file:
            print(code_str, file=print_modified_code_file)

        compile_dict = {
                "_MODULE_SOURCE_CODE": code_str,
                "_set_and_return": _set_and_return,
                }

        # {{{ gather symbols from modules from which methods were collected

        from importlib import import_module
        for mod_name in seen_module_names:
            for name, value in import_module(mod_name).__dict__.items():
                if name.startswith("__"):
                    continue
                if name in compile_dict:
                    if compile_dict[name] is not value:
                        raise ValueError(
                                "symbol disagreement in environment: "
                                f"'{name}', most recently from '{mod_name}'")
                compile_dict[name] = value

        # }}}

        exec(compile(
            code_str,
            f"<'{_get_file_name_for_module_name(cls.__module__)}' "
            "modified by optimize_mapper>",
            "exec"),
             compile_dict)

        return cast(type, compile_dict[cls.__name__])

    return wrapper

# vim: foldmethod=marker



from importlib import metadata
from urllib.request import urlopen


_conf_url = \
        "https://raw.githubusercontent.com/inducer/sphinxconfig/main/sphinxconfig.py"
with urlopen(_conf_url) as _inf:
    exec(compile(_inf.read(), _conf_url, "exec"), globals())

copyright = "2013-24, Andreas Kloeckner"
release = metadata.version("pymbolic")
version = ".".join(release.split(".")[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

intersphinx_mapping = {
    "galgebra": ("https://galgebra.readthedocs.io/en/latest/", None),
    "mako": ("https://docs.makotemplates.org/en/latest/", None),
    "matchpy": ("https://matchpy.readthedocs.io/en/latest/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3", None),
    "sympy": ("https://docs.sympy.org/dev/", None),
    "typing_extensions":
        ("https://typing-extensions.readthedocs.io/en/latest/", None),
    "immutabledict":
        ("https://immutabledict.corenting.fr/", None)
}
autodoc_type_aliases = {
    "Expression": "Expression",
    "ArithmeticExpression": "ArithmeticExpression",
}


nitpick_ignore_regex = [
    # Avoids this error. Not sure where to even look.
    # <unknown>:1: WARNING: py:class reference target not found: ExpressionNode [ref.class]  # noqa: E501
    ["py:class", r"ExpressionNode"],
    ]


import sys


sys._BUILDING_SPHINX_DOCS = True



from __future__ import annotations


__copyright__ = """
Copyright (C) 2017 Matt Wala
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import symengine

import pymbolic.primitives as prim
from pymbolic.interop.common import PymbolicToSympyLikeMapper, SympyLikeToPymbolicMapper


__doc__ = """
.. class:: SymEngineToPymbolicMapper

    .. method:: __call__(expr)

.. class:: PymbolicToSymEngineMapper

    .. method:: __call__(expr)
"""


# {{{ symengine -> pymbolic

class SymEngineToPymbolicMapper(SympyLikeToPymbolicMapper):

    def map_Pow(self, expr):  # noqa
        # SymEngine likes to use as e**a to express exp(a); we undo that here.
        base, exp = expr.args
        if base == symengine.E:
            return prim.Variable("exp")(self.rec(exp))
        else:
            return prim.Power(self.rec(base), self.rec(exp))

    def map_Constant(self, expr):  # noqa
        return self.rec(expr.n())

    map_Complex = map_Constant  # noqa: N815

    def map_ComplexDouble(self, expr):  # noqa
        return complex(expr)

    map_RealDouble = SympyLikeToPymbolicMapper.to_float  # noqa: N815

    def map_Piecewise(self, expr):  # noqa
        # We only handle piecewises with 2 statements!
        if not len(expr.args) == 4:
            raise NotImplementedError
        # We only handle if/else cases
        if not (expr.args[3].is_Boolean and bool(expr.args[3]) is True):
            raise NotImplementedError
        rec_args = [self.rec(arg) for arg in expr.args[:3]]
        then, cond, else_ = rec_args
        return prim.If(cond, then, else_)

    def function_name(self, expr):
        try:
            # For FunctionSymbol instances
            return expr.get_name()
        except AttributeError:
            # For builtin functions
            return type(expr).__name__

    def not_supported(self, expr):
        from symengine.lib.symengine_wrapper import PyFunction  # pylint: disable=E0611
        if isinstance(expr, PyFunction) and \
                self.function_name(expr) == "CSE":       # pylint: disable=E0611
            sympy_expr = expr._sympy_()
            return prim.CommonSubexpression(
                self.rec(expr.args[0]), sympy_expr.prefix, sympy_expr.scope)
        elif isinstance(expr, symengine.Function) and \
                self.function_name(expr) == "CSE":
            return prim.CommonSubexpression(
                self.rec(expr.args[0]), scope=prim.cse_scope.EVALUATION)
        return SympyLikeToPymbolicMapper.not_supported(self, expr)

# }}}


# {{{ pymbolic -> symengine

class PymbolicToSymEngineMapper(PymbolicToSympyLikeMapper):

    sym = symengine

    def raise_conversion_error(self, expr):
        raise RuntimeError(f"do not know how to translate '{expr}' to symengine")


# }}}


CSE = symengine.Function("CSE")


def make_cse(arg, prefix=None, scope=None):
    # SymEngine's classes can't be inherited, but there's a
    # mechanism to create one based on SymPy's ones which stores
    # the SymPy object inside the C++ object.
    # This SymPy object is later retrieved to get the prefix
    # These conversions between SymPy and SymEngine are expensive,
    # so use it only if necessary.
    if prefix is None and scope is None:
        return CSE(arg)
    from pymbolic.interop.sympy import make_cse as make_cse_sympy
    sympy_result = make_cse_sympy(arg, prefix=prefix, scope=scope)
    return symengine.sympify(sympy_result)

# vim: fdm=marker



from __future__ import annotations


__copyright__ = """
Copyright (C) 2015 Andreas Kloeckner
Copyright (C) 2022 Kaushik Kulkarni
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import ast
from typing import Any, ClassVar

import pymbolic.primitives as p
from pymbolic.mapper import CachedMapper
from pymbolic.typing import Expression


__doc__ = r'''

An example::

    src = """
    def f():
        xx = 3*y + z * (12 if x < 13 else 13)
        yy = f(x, y=y)
    """

    import ast
    mod = ast.parse(src.replace("\n    ", "\n"))

    print(ast.dump(mod))

    from pymbolic.interop.ast import ASTToPymbolic
    ast2p = ASTToPymbolic()

    for f in mod.body:
        if not isinstance(f, ast.FunctionDef):
            continue

        for stmt in f.body:
            if not isinstance(stmt, ast.Assign):
                continue

            lhs, = stmt.targets
            lhs = ast2p(lhs)
            rhs = ast2p(stmt.value)

            print(lhs, rhs)

.. autoclass:: ASTToPymbolic
'''


class ASTMapper:
    def __call__(self, expr, *args, **kwargs):
        return self.rec(expr, *args, **kwargs)

    def rec(self, expr, *args, **kwargs):
        mro = list(type(expr).__mro__)
        dispatch_class = kwargs.pop("dispatch_class", type(self))

        while mro:
            method_name = "map_"+mro.pop(0).__name__

            try:
                method = getattr(dispatch_class, method_name)
            except AttributeError:
                pass
            else:
                return method(self, expr, *args, **kwargs)

        return self.not_supported(expr)

    def not_supported(self, expr):
        raise NotImplementedError(
                "{} does not know how to map type '{}'".format(
                    type(self).__name__,
                    type(expr).__name__))


# {{{ mapper

def _add(x, y):
    return p.Sum((x, y))


def _sub(x, y):
    return p.Sum((x, p.Product(((-1), y))))


def _mult(x, y):
    return p.Product((x, y))


def _neg(x):
    return -x


class ASTToPymbolic(ASTMapper):

    bin_op_map: ClassVar[dict[type[ast.operator], Any]] = {
            ast.Add: _add,
            ast.Sub: _sub,
            ast.Mult: _mult,
            # MatMult
            ast.Div: p.Quotient,
            ast.FloorDiv: p.FloorDiv,
            ast.Mod: p.Remainder,
            ast.Pow: p.Power,
            ast.LShift: p.LeftShift,
            ast.RShift: p.RightShift,
            ast.BitOr: p.BitwiseOr,
            ast.BitXor: p.BitwiseXor,
            ast.BitAnd: p.BitwiseAnd,
            }

    def map_BinOp(self, expr):  # noqa
        try:
            op_constructor = self.bin_op_map[type(expr.op)]
        except KeyError:
            raise NotImplementedError(
                f"{type(self).__name__} does not know how to map operator "
                f"'{type(expr.op).__name__}'") from None

        return op_constructor(self.rec(expr.left), self.rec(expr.right))

    unary_op_map: ClassVar[dict[type[ast.unaryop], Any]] = {
            ast.Invert: _neg,
            ast.Not: p.LogicalNot,
            # ast.UAdd:
            ast.USub: _neg,
            }

    def map_UnaryOp(self, expr):  # noqa
        try:
            op_constructor = self.unary_op_map[type(expr.op)]
        except KeyError:
            raise NotImplementedError(
                f"{type(self).__name__} does not know how to map operator "
                f"'{type(expr.op).__name__}'") from None

        return op_constructor(self.rec(expr.operand))

    def map_IfExp(self, expr):  # noqa
        # (expr test, expr body, expr orelse)
        return p.If(self.rec(expr.test), self.rec(expr.body), self.rec(expr.orelse))

    comparison_op_map: ClassVar[dict[type[ast.cmpop], str]] = {
            ast.Eq: "==",
            ast.NotEq: "!=",
            ast.Lt: "<",
            ast.LtE: "<=",
            ast.Gt: ">",
            ast.GtE: ">=",
            # Is
            # IsNot
            # In
            # NotIn
            }

    def map_Compare(self, expr):  # noqa
        # (expr left, cmpop* ops, expr* comparators)
        op, = expr.ops

        try:
            comp = self.comparison_op_map[type(op)]
        except KeyError:
            raise NotImplementedError(
                f"{type(self).__name__} does not know how to map operator "
                f"'{type(expr.op).__name__}'") from None

        # FIXME: Support strung-together comparisons
        right, = expr.comparators

        return p.Comparison(self.rec(expr.left), comp, self.rec(right))

    def map_Call(self, expr):  # noqa
        # (expr func, expr* args, keyword* keywords)
        func = self.rec(expr.func)
        args = tuple([self.rec(arg) for arg in expr.args])
        if getattr(expr, "keywords", []):
            return p.CallWithKwargs(func, args,
                    {
                        kw.arg: self.rec(kw.value)
                        for kw in expr.keywords})
        else:
            return p.Call(func, args)

    def map_Num(self, expr):  # noqa
        # (object n) -- a number as a PyObject.
        return expr.n

    def map_Str(self, expr):  # noqa
        return expr.s

    def map_Bytes(self, expr):  # noqa
        return expr.s

    def map_Constant(self, expr):  # noqa
        # (singleton value)
        return expr.value

    def map_Attribute(self, expr):  # noqa
        # (expr value, identifier attr, expr_context ctx)
        return p.Lookup(self.rec(expr.value), expr.attr)

    def map_Subscript(self, expr):  # noqa
        # (expr value, slice slice, expr_context ctx)
        def none_or_rec(x):
            if x is None:
                return x
            else:
                return self.rec(x)

        if isinstance(expr.slice, slice):
            index = slice(
                    none_or_rec(expr.slice.start),
                    none_or_rec(expr.slice.stop),
                    none_or_rec(expr.slice.step))
        else:
            index = none_or_rec(expr.slice)

        return p.Subscript(
                self.rec(expr.value),
                index)

    # def map_Starred(self, expr):

    def map_Name(self, expr):  # noqa
        # (identifier id, expr_context ctx)
        return p.Variable(expr.id)

    def map_Tuple(self, expr):  # noqa
        # (expr* elts, expr_context ctx)
        return tuple([self.rec(ti) for ti in expr.elts])

# }}}


# {{{ PymbolicToASTMapper

class PymbolicToASTMapper(CachedMapper[ast.expr, []]):
    def map_variable(self, expr) -> ast.expr:
        return ast.Name(id=expr.name)

    def _map_multi_children_op(self,
                               children: tuple[Expression, ...],
                               op_type: ast.operator) -> ast.expr:
        rec_children = [self.rec(child) for child in children]
        result = rec_children[-1]
        for child in rec_children[-2::-1]:
            result = ast.BinOp(child, op_type, result)

        return result

    def map_sum(self, expr: p.Sum) -> ast.expr:
        return self._map_multi_children_op(expr.children, ast.Add())

    def map_product(self, expr: p.Product) -> ast.expr:
        return self._map_multi_children_op(expr.children, ast.Mult())

    def map_constant(self, expr: object) -> ast.expr:
        return ast.Constant(expr, None)

    def map_call(self, expr: p.Call) -> ast.expr:
        return ast.Call(
            func=self.rec(expr.function),
            args=[self.rec(param) for param in expr.parameters],
            keywords=[],
        )

    def map_call_with_kwargs(self, expr) -> ast.expr:
        return ast.Call(
            func=self.rec(expr.function),
            args=[self.rec(param) for param in expr.parameters],
            keywords=[
                ast.keyword(
                    arg=kw,
                    value=self.rec(param))
                for kw, param in sorted(expr.kw_parameters.items())])

    def map_subscript(self, expr) -> ast.expr:
        return ast.Subscript(value=self.rec(expr.aggregate),
                             slice=self.rec(expr.index))

    def map_lookup(self, expr) -> ast.expr:
        return ast.Attribute(self.rec(expr.aggregate),
                             expr.name)

    def map_quotient(self, expr) -> ast.expr:
        return self._map_multi_children_op((expr.numerator,
                                            expr.denominator),
                                           ast.Div())

    def map_floor_div(self, expr) -> ast.expr:
        return self._map_multi_children_op((expr.numerator,
                                            expr.denominator),
                                           ast.FloorDiv())

    def map_remainder(self, expr) -> ast.expr:
        return self._map_multi_children_op((expr.numerator,
                                            expr.denominator),
                                           ast.Mod())

    def map_power(self, expr) -> ast.expr:
        return self._map_multi_children_op((expr.base,
                                            expr.exponent),
                                           ast.Pow())

    def map_left_shift(self, expr) -> ast.expr:
        return self._map_multi_children_op((expr.shiftee,
                                            expr.shift),
                                           ast.LShift())

    def map_right_shift(self, expr) -> ast.expr:
        return self._map_multi_children_op((expr.numerator,
                                            expr.denominator),
                                           ast.RShift())

    def map_bitwise_not(self, expr) -> ast.expr:
        return ast.UnaryOp(ast.Invert(), self.rec(expr.child))

    def map_bitwise_or(self, expr) -> ast.expr:
        return self._map_multi_children_op(expr.children,
                                           ast.BitOr())

    def map_bitwise_xor(self, expr) -> ast.expr:
        return self._map_multi_children_op(expr.children,
                                           ast.BitXor())

    def map_bitwise_and(self, expr) -> ast.expr:
        return self._map_multi_children_op(expr.children,
                                           ast.BitAnd())

    def map_logical_not(self, expr) -> ast.expr:
        return ast.UnaryOp(ast.Not(), self.rec(expr.child))

    def map_logical_or(self, expr) -> ast.expr:
        return ast.BoolOp(ast.Or(), [self.rec(child)
                                     for child in expr.children])

    def map_logical_and(self, expr) -> ast.expr:
        return ast.BoolOp(ast.And(), [self.rec(child)
                                     for child in expr.children])

    def map_list(self, expr: list[Any]) -> ast.expr:
        return ast.List([self.rec(el) for el in expr])

    def map_tuple(self, expr: tuple[Any, ...]) -> ast.expr:
        return ast.Tuple([self.rec(el) for el in expr])

    def map_if(self, expr: p.If) -> ast.expr:
        return ast.IfExp(test=self.rec(expr.condition),
                         body=self.rec(expr.then),
                         orelse=self.rec(expr.else_))

    def map_nan(self, expr: p.NaN) -> ast.expr:
        assert expr.data_type is not None
        if isinstance(expr.data_type(float("nan")), float):
            return ast.Call(
                ast.Name(id="float"),
                args=[ast.Constant("nan")],
                keywords=[])
        else:
            # TODO: would need attributes of NumPy
            raise NotImplementedError("Non-float nan not implemented")

    def map_slice(self, expr: p.Slice) -> ast.expr:
        return ast.Slice(*[None if child is None else self.rec(child)
                           for child in expr.children])

    def map_numpy_array(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_multivector(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_common_subexpression(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_substitution(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_derivative(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_if_positive(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_comparison(self, expr: p.Comparison) -> ast.expr:
        raise NotImplementedError

    def map_wildcard(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_dot_wildcard(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_star_wildcard(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_function_symbol(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_min(self, expr) -> ast.expr:
        raise NotImplementedError

    def map_max(self, expr) -> ast.expr:
        raise NotImplementedError


def to_python_ast(expr) -> ast.expr:
    """
    Maps *expr* to :class:`ast.expr`.
    """
    return PymbolicToASTMapper()(expr)


def to_evaluatable_python_function(expr: Expression,
                                   fn_name: str
                                   ) -> str:
    """
    Returns a :class:`str` of the python code with a single function *fn_name*
    that takes in the variables in *expr* as keyword-only arguments and returns
    the evaluated value of *expr*.

    .. testsetup::

        >>> from pymbolic import parse
        >>> from pymbolic.interop.ast import to_evaluatable_python_function

    .. doctest::

        >>> expr = parse("S//32 + E%32")
        >>> print(to_evaluatable_python_function(expr, "foo"))
        def foo(*, E, S):
            return S // 32 + E % 32
    """

    from pymbolic.mapper.dependency import CachedDependencyMapper

    dep_mapper: CachedDependencyMapper[[]] = (
        CachedDependencyMapper(composite_leaves=True))

    deps: list[str] = []
    for dep in dep_mapper(expr):
        if isinstance(dep, p.Variable):
            deps.append(dep.name)
        else:
            raise NotImplementedError(f"{dep!r} is not supported")

    ast_func = ast.FunctionDef(name=fn_name,
                               args=ast.arguments(args=[],
                                                  posonlyargs=[],
                                                  kwonlyargs=[ast.arg(dep, None)
                                                              for dep in sorted(deps)],
                                                  kw_defaults=[None]*len(deps),
                                                  vararg=None,
                                                  kwarg=None,
                                                  defaults=[]),
                               body=[ast.Return(to_python_ast(expr))],
                               decorator_list=[])
    ast_module = ast.Module([ast_func], type_ignores=[])

    return ast.unparse(ast.fix_missing_locations(ast_module))

# }}}

# vim: foldmethod=marker



"""
.. autoclass:: DifferentiationMapper
"""
from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import pymbolic
import pymbolic.mapper
import pymbolic.mapper.evaluator
import pymbolic.primitives as primitives


def map_math_functions_by_name(i, func, pars, allowed_nonsmoothness="none"):
    def make_f(name):
        return primitives.Lookup(primitives.Variable("math"), name)

    if func == make_f("sin") and len(pars) == 1:
        return make_f("cos")(*pars)
    elif func == make_f("cos") and len(pars) == 1:
        return -make_f("sin")(*pars)
    elif func == make_f("tan") and len(pars) == 1:
        return make_f("tan")(*pars)**2+1
    elif func == make_f("log") and len(pars) == 1:
        return primitives.quotient(1, pars[0])
    elif func == make_f("exp") and len(pars) == 1:
        return make_f("exp")(*pars)
    elif func == make_f("sinh") and len(pars) == 1:
        return make_f("cosh")(*pars)
    elif func == make_f("cosh") and len(pars) == 1:
        return make_f("sinh")(*pars)
    elif func == make_f("tanh") and len(pars) == 1:
        return 1-make_f("tanh")(*pars)**2
    elif func == make_f("expm1") and len(pars) == 1:
        return make_f("exp")(*pars)
    elif func == make_f("fabs") and len(pars) == 1:
        if allowed_nonsmoothness in ["continuous", "discontinuous"]:
            from pymbolic.functions import sign
            return sign(*pars)
        else:
            raise ValueError("fabs is not smooth"
                             ", pass allowed_nonsmoothness='continuous' "
                             "to return sign")
    elif func == make_f("copysign") and len(pars) == 2:
        if allowed_nonsmoothness == "discontinuous":
            return 0
        else:
            raise ValueError("sign is discontinuous"
                             ", pass allowed_nonsmoothness='discontinuous' "
                             "to return 0")
    else:
        raise RuntimeError("unrecognized function, cannot differentiate")


class DifferentiationMapper(pymbolic.mapper.Mapper,
        pymbolic.mapper.CSECachingMapperMixin):
    """Example usage:

    .. doctest::
        :options: +NORMALIZE_WHITESPACE

        >>> import pymbolic.primitives as p
        >>> x = p.Variable("x")
        >>> expr = x*(x+5)**3/(x-1)**2

        >>> from pymbolic import flatten
        >>> from pymbolic.mapper.differentiator import DifferentiationMapper as DM
        >>> print(flatten(DM(x)(expr)))
        (((x + 5)**3 + x*3*(x + 5)**2)*(x + -1)**2 + (-1)*2*(x + -1)*x*(x + 5)**3) / (x + -1)**2**2
    """  # noqa: E501

    def __init__(self, variable, func_map=map_math_functions_by_name,
                 allowed_nonsmoothness=None):
        """
        :arg variable: A :class:`pymbolic.primitives.Variable` instance
            by which to differentiate.
        :arg func_map: A function for computing derivatives of function
            calls, signature ``(arg_index, function_variable, parameters)``.
        :arg allowed_nonsmoothness: Whether to allow differentiation of
            functions which are not smooth or continuous.
            Pass ``"continuous"`` to allow nonsmooth but not discontinuous
            functions or ``"discontinuous"`` to allow both.
            Defaults to ``"none"``, in which case neither is allowed.

        .. versionchanged:: 2019.2

            Added *allowed_nonsmoothness*.
        """

        if allowed_nonsmoothness is None:
            allowed_nonsmoothness = "none"

        self.variable = variable
        self.function_map = func_map
        if allowed_nonsmoothness not in ["none", "continuous", "discontinuous"]:
            raise ValueError(f"allowed_nonsmoothness={allowed_nonsmoothness} "
                    "is not a valid option")
        self.allowed_nonsmoothness = allowed_nonsmoothness

    def rec_undiff(self, expr, *args):
        """This method exists for the benefit of subclasses that may need to
        process un-differentiated subexpressions.
        """
        return expr

    def map_constant(self, expr, *args):
        return 0

    def map_variable(self, expr, *args):
        if expr == self.variable:
            return 1
        else:
            return 0

    def map_call(self, expr, *args):
        return pymbolic.flattened_sum(
            self.function_map(
                i, expr.function, self.rec_undiff(expr.parameters, *args),
                allowed_nonsmoothness=self.allowed_nonsmoothness)
            * self.rec(par, *args)
            for i, par in enumerate(expr.parameters)
            )

    map_subscript = map_variable

    def map_sum(self, expr, *args):
        return pymbolic.flattened_sum(
                self.rec(child, *args) for child in expr.children)

    def map_product(self, expr, *args):
        return pymbolic.flattened_sum(
            pymbolic.flattened_product(
                [self.rec_undiff(ch, *args) for ch in expr.children[0:i]]
                + [self.rec(child, *args)]
                + [self.rec_undiff(ch, *args) for ch in expr.children[i+1:]]
                )
            for i, child in enumerate(expr.children))

    def map_quotient(self, expr, *args):
        f = expr.numerator
        g = expr.denominator
        df = self.rec(f, *args)
        dg = self.rec(g, *args)
        f = self.rec_undiff(f, *args)
        g = self.rec_undiff(g, *args)

        if (not df) and (not dg):
            return 0
        elif (not df):
            return -f*dg/g**2
        elif (not dg):
            return self.rec(f, *args)/g
        else:
            return (df*g-dg*f)/g**2

    def map_power(self, expr, *args):
        f = expr.base
        g = expr.exponent
        df = self.rec(f, *args)
        dg = self.rec(g, *args)
        f = self.rec_undiff(f, *args)
        g = self.rec_undiff(g, *args)

        log = pymbolic.var("log")

        if (not df) and (not dg):
            return 0
        elif (not df):
            return log(f) * f**g * dg
        elif (not dg):
            return g * f**(g-1) * df
        else:
            return log(f) * f**g * dg + \
                    g * f**(g-1) * df

    def map_numpy_array(self, expr, *args):
        import numpy
        result = numpy.empty(expr.shape, dtype=object)
        for i in numpy.ndindex(result.shape):
            result[i] = self.rec(expr[i], *args)
        return result

    def map_if(self, expr, *args):
        if self.allowed_nonsmoothness != "discontinuous":
            raise ValueError("cannot differentiate 'If' nodes unless "
                    "allowed_nonsmoothness is set to 'discontinuous'")

        return type(expr)(
                expr.condition,
                self.rec(expr.then, *args),
                self.rec(expr.else_, *args))

    def map_common_subexpression_uncached(self, expr, *args):
        return type(expr)(
                self.rec(expr.child, *args),
                expr.prefix,
                expr.scope)


def differentiate(expression,
                  variable,
                  func_mapper=map_math_functions_by_name,
                  allowed_nonsmoothness="none"):
    if not isinstance(variable, primitives.Variable | primitives.Subscript):
        variable = primitives.make_variable(variable)
    from pymbolic import flatten
    return flatten(DifferentiationMapper(
        variable, func_mapper, allowed_nonsmoothness=allowed_nonsmoothness
        )(expression))



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Generic, TypeVar, cast

import numpy as np

from pytools import memoize, memoize_method

from pymbolic.primitives import expr_dataclass, is_zero
from pymbolic.typing import ArithmeticExpression, T


__doc__ = """
See `Wikipedia <https://en.wikipedia.org/wiki/Geometric_algebra>`__ for an idea
of what this is.

.. versionadded:: 2013.2

Also see :ref:`ga-examples`.

Spaces
------

.. autoclass:: Space

.. autofunction:: get_euclidean_space

Multivectors
------------

.. class:: CoeffT

    A type variable for coefficients of :class:`MultiVector`. Requires some arithmetic.

.. autoclass:: MultiVector

.. _ga-examples:

Example usage
-------------

This first example demonstrates how to compute a cross product using
:class:`MultiVector`:

.. doctest::

    >>> import numpy as np
    >>> import pymbolic.geometric_algebra as ga
    >>> MV = ga.MultiVector

    >>> a = np.array([3.344, 1.2, -0.5])
    >>> b = np.array([7.4, 1.1, -2.0])
    >>> np.cross(a, b)
    array([-1.85  ,  2.988 , -5.2016])

    >>> mv_a = MV(a)
    >>> mv_b = MV(b)
    >>> print(-mv_a.I*(mv_a^mv_b))
    MV(
        e0 * -1.8499999999999999
        + e1 * 2.9879999999999995
        + e2 * -5.201600000000001)

This simple example demonstrates how a complex number is simply a special
case of a :class:`MultiVector`:

.. doctest::

    >>> import numpy as np
    >>> import pymbolic.geometric_algebra as ga
    >>> MV = ga.MultiVector
    >>>
    >>> sp = ga.Space(metric_matrix=-np.eye(1))
    >>> sp
    Space(['e0'], array([[-1.]]))

    >>> one = MV(1, sp)
    >>> one
    MultiVector({0: 1}, Space(['e0'], array([[-1.]])))
    >>> print(one)
    MV(1)
    >>> print(one.I)
    MV(e0 * 1)
    >>> print(one.I ** 2)
    MV(-1.0)

    >>> print((3+5j)*(2+3j)/(3j))
    (6.333333333333333+3j)
    >>> print((3+5*one.I)*(2+3*one.I)/(3*one.I))
    MV(
        6.333333333333333
        + e0 * 3.0)

The following test demonstrates the use of the object and shows many useful
properties:

.. literalinclude:: ../test/test_pymbolic.py
   :start-after: START_GA_TEST
   :end-before: END_GA_TEST

"""


# {{{ helpers

def permutation_sign(p: Iterable[int]) -> int:
    """
    :returns: the sign of the permutation *p*.
    """
    p = list(p)
    s = +1

    for i in range(len(p)):
        # j is the current position of item I.
        j = i

        while p[j] != i:
            j += 1

        # Unless the item is already in the correct place, restore it.
        if j != i:
            p[i], p[j] = p[j], p[i]
            s = -s

    return s


def canonical_reordering_sign(a_bits: int, b_bits: int) -> int:
    """Count the number of basis vector swaps required to
    get the combination of *a_bits* and *b_bits* into canonical order.

    Algorithm from figure 19.1 of [DFM2010]_ in :class:`MultiVector`.

    :arg a_bits: bitmap representing basis blade *a*.
    :arg b_bits: bitmap representing basis blade *b*.
    """

    a_bits = a_bits >> 1
    s = 0
    while a_bits:
        s = s + (a_bits & b_bits).bit_count()
        a_bits = a_bits >> 1

    if s & 1:
        return -1
    else:
        return 1

# }}}


# {{{ space

@dataclass(frozen=True, init=False)
class Space:
    """
    .. autoattribute :: basis_names
    .. autoattribute :: metric_matrix

    .. autoproperty :: dimensions
    .. autoproperty :: is_orthogonal
    .. autoproperty :: is_euclidean

    .. automethod:: bits_and_sign
    .. automethod:: blade_bits_to_str
    """

    basis_names: Sequence[str]
    "A sequence of names of basis vectors."

    metric_matrix: np.ndarray
    """
    A *(dims, dims)*-shaped matrix, whose *(i, j)*-th entry represents the
    inner product of basis vector *i* and basis vector *j*.
    """

    def __init__(self,
                 basis: Sequence[str] | int | None = None,
                 metric_matrix: np.ndarray | None = None) -> None:
        """
        :arg basis: A sequence of names of basis vectors, or an integer (the
            number of dimensions) to use the default names ``e0`` through ``eN``.
        :arg metric_matrix: See :attr:`metric_matrix`. If *None*, the Euclidean
            metric is assumed.
        """

        if basis is None and metric_matrix is None:
            raise TypeError(
                "At least one of 'basis' and 'metric_matrix' must be given")

        from numbers import Integral
        if basis is None:
            assert metric_matrix is not None
            basis_names = [f"e{i}" for i in range(metric_matrix.shape[0])]
        elif isinstance(basis, Integral):
            basis_names = [f"e{i}" for i in range(basis)]
        else:
            assert not isinstance(basis, int)
            basis_names = list(basis)

        if metric_matrix is None:
            metric_matrix = np.eye(len(basis_names), dtype=object)

        if not (
                len(metric_matrix.shape) == 2
                and all(dim == len(basis_names) for dim in metric_matrix.shape)):
            raise ValueError(
                f"'metric_matrix' has the wrong shape: {metric_matrix.shape}")

        object.__setattr__(self, "basis_names", basis_names)
        object.__setattr__(self, "metric_matrix", metric_matrix)

    @property
    def dimensions(self) -> int:
        """The dimension of the space."""
        return len(self.basis_names)

    @memoize_method
    def bits_and_sign(self, basis_indices: Sequence[int]) -> tuple[int, int]:
        # assert no repetitions
        assert len(set(basis_indices)) == len(basis_indices)

        sorted_basis_indices = tuple(sorted(
                (bindex, num)
                for num, bindex in enumerate(basis_indices)))
        blade_permutation = [num for _, num in sorted_basis_indices]

        bits = 0
        for bi in basis_indices:
            bits |= 2**bi

        return bits, permutation_sign(blade_permutation)

    @property
    @memoize_method
    def is_orthogonal(self):
        """*True* if the metric is orthogonal (i.e. diagonal)."""
        return (self.metric_matrix - np.diag(np.diag(self.metric_matrix)) == 0).all()

    @property
    @memoize_method
    def is_euclidean(self) -> bool:
        """*True* if the metric matrix corresponds to the Eucledian metric."""
        return (self.metric_matrix == np.eye(self.metric_matrix.shape[0])).all()

    def blade_bits_to_str(self, bits: int, outer_operator: str = "^") -> str:
        return outer_operator.join(
                    name
                    for bit_num, name in enumerate(self.basis_names)
                    if bits & (1 << bit_num))

    def __repr__(self) -> str:
        if self is get_euclidean_space(self.dimensions):
            return f"Space({self.dimensions})"

        elif self.is_euclidean:
            return f"Space({self.basis_names!r})"
        else:
            return f"Space({self.basis_names!r}, {self.metric_matrix!r})"


@memoize
def get_euclidean_space(n: int) -> Space:
    """Return the canonical *n*-dimensional Euclidean :class:`Space`."""
    return Space(n)

# }}}


CoeffT = TypeVar("CoeffT", bound=ArithmeticExpression)


# {{{ blade product weights

def _shared_metric_coeff(shared_bits, space):
    result = 1

    basis_idx = 0
    while shared_bits:
        bit = (1 << basis_idx)
        if shared_bits & bit:
            result = result * space.metric_matrix[basis_idx, basis_idx]
            shared_bits ^= bit

        basis_idx += 1

    return result


class _GAProduct(ABC, Generic[CoeffT]):
    @staticmethod
    @abstractmethod
    def generic_blade_product_weight(a_bits: int, b_bits: int, space: Space) -> CoeffT:
        ...

    @staticmethod
    @abstractmethod
    def orthogonal_blade_product_weight(
                a_bits: int, b_bits: int, space: Space
            ) -> CoeffT:
        ...


class _OuterProduct(_GAProduct):
    @staticmethod
    def generic_blade_product_weight(a_bits, b_bits, space):
        return int(not a_bits & b_bits)

    orthogonal_blade_product_weight = generic_blade_product_weight


class _GeometricProduct(_GAProduct):
    @staticmethod
    def generic_blade_product_weight(a_bits, b_bits, space):
        raise NotImplementedError("geometric product for spaces "
                "with non-diagonal metric (i.e. non-orthogonal basis)")

    @staticmethod
    def orthogonal_blade_product_weight(a_bits, b_bits, space):
        shared_bits = a_bits & b_bits

        if shared_bits:
            return _shared_metric_coeff(shared_bits, space)
        else:
            return 1


class _InnerProduct(_GAProduct):
    @staticmethod
    def generic_blade_product_weight(a_bits, b_bits, space):
        raise NotImplementedError("inner product for spaces "
                "with non-diagonal metric (i.e. non-orthogonal basis)")

    @staticmethod
    def orthogonal_blade_product_weight(a_bits, b_bits, space):
        shared_bits = a_bits & b_bits

        if shared_bits == a_bits or shared_bits == b_bits:
            return _shared_metric_coeff(shared_bits, space)
        else:
            return 0


class _LeftContractionProduct(_GAProduct):
    @staticmethod
    def generic_blade_product_weight(a_bits, b_bits, space):
        raise NotImplementedError("contraction product for spaces "
                "with non-diagonal metric (i.e. non-orthogonal basis)")

    @staticmethod
    def orthogonal_blade_product_weight(a_bits, b_bits, space):
        shared_bits = a_bits & b_bits

        if shared_bits == a_bits:
            return _shared_metric_coeff(shared_bits, space)
        else:
            return 0


class _RightContractionProduct(_GAProduct):
    @staticmethod
    def generic_blade_product_weight(a_bits, b_bits, space):
        raise NotImplementedError("contraction product for spaces "
                "with non-diagonal metric (i.e. non-orthogonal basis)")

    @staticmethod
    def orthogonal_blade_product_weight(a_bits, b_bits, space):
        shared_bits = a_bits & b_bits

        if shared_bits == b_bits:
            return _shared_metric_coeff(shared_bits, space)
        else:
            return 0


class _ScalarProduct(_GAProduct):
    @staticmethod
    def generic_blade_product_weight(a_bits, b_bits, space):
        raise NotImplementedError("contraction product for spaces "
                "with non-diagonal metric (i.e. non-orthogonal basis)")

    @staticmethod
    def orthogonal_blade_product_weight(a_bits, b_bits, space):
        if a_bits == b_bits:
            return _shared_metric_coeff(a_bits, space)
        else:
            return 0

# }}}


# {{{ multivector

def _cast_to_mv(obj: Any, space: Space) -> MultiVector:
    if isinstance(obj, MultiVector):
        return obj
    else:
        return MultiVector(obj, space)


@expr_dataclass(init=False, hash=False)
class MultiVector(Generic[CoeffT]):
    r"""An immutable multivector type. Its implementation follows [DFM].
    It is pickleable, and not picky about what data is used as coefficients.
    It supports :class:`pymbolic.primitives.ExpressionNode` objects of course,
    but it can take just about any other scalar-ish coefficients.

    .. autoattribute:: data

    .. autoattribute:: space

    See the following literature:

        [DFM] L. Dorst, D. Fontijne, and S. Mann, `Geometric Algebra for Computer
        Science: An Object-Oriented Approach to Geometry
        <https://books.google.com?isbn=0080553109>`__. Morgan Kaufmann, 2010.

        [HS] D. Hestenes and G. Sobczyk, `Clifford Algebra to Geometric Calculus: A
        Unified Language for Mathematics and Physics
        <https://books.google.com?isbn=9027725616>`__. Springer, 1987.

    The object behaves much like the corresponding :class:`galgebra.mv.Mv`
    object in :mod:`galgebra`, especially with respect to the supported
    operators:

    =================== ========================================================
    Operation           Result
    =================== ========================================================
    ``A+B``             Sum of multivectors
    ``A-B``             Difference of multivectors
    ``A*B``             Geometric product :math:`AB`
    ``A^B``             Outer product :math:`A\wedge B` of multivectors
    ``A|B``             Inner product :math:`A\cdot B` of multivectors
    ``A<<B``            Left contraction :math:`A\lrcorner B` (``_|``)
                        of multivectors, also read as ':math:`A` removed from
                        :math:`B`'.
    ``A>>B``            Right contraction :math:`A\llcorner B`  (``|_``)of
                        multivectors, also read as ':math:`A` without
                        :math:`B`'.
    =================== ========================================================

    .. warning ::

        Many of the multiplicative operators bind more weakly than
        even *addition*. Python's operator precedence further does not
        match geometric algebra, which customarily evaluates outer, inner,
        and then geometric.

        In other words: Use parentheses everywhere.

    .. autoattribute:: mapper_method

    .. rubric:: More products

    .. automethod:: scalar_product
    .. automethod:: x
    .. automethod:: __pow__

    .. rubric:: Unary operators

    .. automethod:: inv
    .. automethod:: rev
    .. automethod:: invol
    .. automethod:: dual
    .. automethod:: __inv__
    .. automethod:: norm_squared
    .. automethod:: __abs__
    .. autoattribute:: I

    .. rubric:: Comparisons

    :class:`MultiVector` objects have a truth value corresponding to whether
    they have any blades with non-zero coefficients. They support testing
    for (exact) equality.

    .. automethod:: zap_near_zeros
    .. automethod:: close_to

    .. rubric:: Grade manipulation

    .. automethod:: gen_blades
    .. automethod:: project
    .. automethod:: xproject
    .. automethod:: all_grades
    .. automethod:: get_pure_grade
    .. automethod:: odd
    .. automethod:: even
    .. automethod:: project_min_grade
    .. automethod:: project_max_grade

    .. automethod:: as_scalar
    .. automethod:: as_vector

    .. rubric:: Helper functions

    .. automethod:: map

    """

    data: Mapping[int, CoeffT]
    """A mapping from a basis vector bitmap (indicating blades) to coefficients.
    (see [DFM], Chapter 19 for idea and rationale)
    """

    space: Space

    mapper_method = "map_multivector"

    # {{{ construction

    def __init__(
                self,
                data: (Mapping[int, CoeffT]
                       | Mapping[tuple[int, ...], CoeffT]
                       | np.ndarray
                       | CoeffT),
                space: Space | None = None
            ) -> None:
        """
        :arg data: This may be one of the following:

            * a :class:`numpy.ndarray`, which will be turned into a grade-1
              multivector,
            * a mapping from tuples of basis indices (together indicating a blade,
              order matters and will be mapped to 'normalized' blades) to
              coefficients,
            * an array as described in :attr:`data`,
            * a scalar--where everything that doesn't fall into the above cases
              is viewed as a scalar.
        :arg space: A :class:`Space` instance. If *None* or an integer,
            :func:`get_euclidean_space` is called to obtain a default space with
            the right number of dimensions for *data*. Note: dimension guessing only
            works when a :class:`numpy.ndarray` is being passed for *data*.
        """

        data_dict: Mapping
        if isinstance(data, np.ndarray):
            if len(data.shape) != 1:
                raise ValueError(
                    "Only numpy vectors (not higher-rank objects) "
                    f"are supported for 'data': shape {data.shape}")

            dimensions, = data.shape
            data_dict = {(i,): cast(CoeffT, xi) for i, xi in enumerate(data)}

            if space is None:
                space = get_euclidean_space(dimensions)

            if space.dimensions != dimensions:
                raise ValueError(
                    "Dimension of 'space' does not match that of 'data': "
                    f"got {space.dimensions}d space but expected {dimensions}d")
        elif isinstance(data, Mapping):
            data_dict = data
        else:
            data_dict = {0: cast(CoeffT, data)}

        if space is None:
            raise ValueError("No 'space' provided")

        # {{{ normalize data to bitmaps, if needed

        from pytools import single_valued

        if data_dict and single_valued(isinstance(k, tuple) for k in data_dict.keys()):
            # data is in non-normalized non-bits tuple form
            new_data: dict[int, CoeffT] = {}
            for basis_indices, coeff in data_dict.items():
                assert isinstance(basis_indices, tuple)

                bits, sign = space.bits_and_sign(basis_indices)
                new_coeff = cast(CoeffT,
                    new_data.setdefault(bits, cast(CoeffT, 0))  # type: ignore[operator]
                    + sign*coeff)

                if is_zero(new_coeff):
                    del new_data[bits]
                else:
                    new_data[bits] = new_coeff
        else:
            new_data = cast(dict[int, CoeffT], data_dict)

        # }}}

        # assert that multivectors don't get nested
        assert not any(isinstance(coeff, MultiVector) for coeff in new_data.values())

        object.__setattr__(self, "space", space)
        object.__setattr__(self, "data", new_data)

    # }}}

    # {{{ stringification

    def stringify(self, coeff_stringifier, enclosing_prec):
        from pymbolic.mapper.stringifier import PREC_PRODUCT, PREC_SUM

        terms = []
        for bits in sorted(self.data.keys(),
                key=lambda _bits: (_bits.bit_count(), _bits)):
            coeff = self.data[bits]

            # {{{ try to find a stringifier

            strifier = None
            if coeff_stringifier is None:
                try:
                    strifier = coeff.stringifier()()
                except AttributeError:
                    pass
            else:
                strifier = coeff_stringifier

            # }}}

            if strifier is not None:
                if bits:
                    coeff_str = strifier(coeff, PREC_PRODUCT)
                else:
                    coeff_str = strifier(coeff, PREC_SUM)
            else:
                coeff_str = str(coeff)

            blade_str = self.space.blade_bits_to_str(bits)
            if bits:
                terms.append(f"{blade_str} * {coeff_str}")
            else:
                terms.append(coeff_str)

        if terms:
            if any(len(t) > 15 for t in terms):
                result = "\n    " + "\n    + ".join(terms)
            else:
                result = " + ".join(terms)
        else:
            result = "0"

        return f"MV({result})"

    def __str__(self):
        from pymbolic.mapper.stringifier import PREC_NONE
        return self.stringify(None, PREC_NONE)

    def __repr__(self):
        return f"MultiVector({self.data}, {self.space!r})"

    # }}}

    # {{{ additive operators

    def __neg__(self) -> MultiVector:
        return MultiVector(
                {bits: -coeff
                    for bits, coeff in self.data.items()},
                self.space)

    def __add__(self, other) -> MultiVector:
        other = _cast_to_mv(other, self.space)

        if self.space is not other.space:
            raise ValueError("can only add multivectors from identical spaces")

        all_bits = set(self.data.keys()) | set(other.data.keys())

        from pymbolic.primitives import is_zero
        new_data = {}
        for bits in all_bits:
            new_coeff = (self.data.get(bits, cast(CoeffT, 0))
                + other.data.get(bits, cast(CoeffT, 0)))

            if not is_zero(new_coeff):
                new_data[bits] = new_coeff

        return MultiVector(new_data, self.space)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    # }}}

    # {{{ multiplicative operators

    def _generic_product(self,
                    other: MultiVector,
                    product_class: _GAProduct
                ) -> MultiVector:
        """
        :arg product_class: A subclass of :class:`_GAProduct`.
        """

        if self.space.is_orthogonal:
            bpw = product_class.orthogonal_blade_product_weight
        else:
            bpw = product_class.generic_blade_product_weight

        if self.space is not other.space:
            raise ValueError("can only compute products of multivectors "
                    "from identical spaces")

        from pymbolic.primitives import is_zero
        new_data: dict[int, CoeffT] = {}
        for sbits, scoeff in self.data.items():
            for obits, ocoeff in other.data.items():
                new_bits = sbits ^ obits
                weight = bpw(sbits, obits, self.space)

                if not is_zero(weight):
                    # These are nonzero by definition.
                    coeff = (weight
                            * canonical_reordering_sign(sbits, obits)
                            * scoeff * ocoeff)
                    new_coeff = new_data.setdefault(new_bits, cast(CoeffT, 0)) + coeff
                    if is_zero(new_coeff):
                        del new_data[new_bits]
                    else:
                        new_data[new_bits] = new_coeff

        return MultiVector(new_data, self.space)

    def __mul__(self, other):
        other = _cast_to_mv(other, self.space)

        return self._generic_product(other, _GeometricProduct)

    def __rmul__(self, other):
        return MultiVector(other, self.space) \
                ._generic_product(self, _GeometricProduct)

    def __xor__(self, other):
        other = _cast_to_mv(other, self.space)

        return self._generic_product(other, _OuterProduct)

    def __rxor__(self, other):
        return MultiVector(other, self.space) \
                ._generic_product(self, _OuterProduct)

    def __or__(self, other):
        other = _cast_to_mv(other, self.space)

        return self._generic_product(other, _InnerProduct)

    def __ror__(self, other):
        return MultiVector(other, self.space)\
                ._generic_product(self, _InnerProduct)

    def __lshift__(self, other):
        other = _cast_to_mv(other, self.space)

        return self._generic_product(other, _LeftContractionProduct)

    def __rlshift__(self, other):
        return MultiVector(other, self.space)\
                ._generic_product(self, _LeftContractionProduct)

    def __rshift__(self, other):
        other = _cast_to_mv(other, self.space)

        return self._generic_product(other, _RightContractionProduct)

    def __rrshift__(self, other):
        return MultiVector(other, self.space)\
                ._generic_product(self, _RightContractionProduct)

    def scalar_product(self, other):
        r"""Return the scalar product, as a scalar, not a :class:`MultiVector`.

        Often written :math:`A*B`.
        """

        other_new = _cast_to_mv(other, self.space)

        return self._generic_product(other_new, _ScalarProduct).as_scalar()

    def x(self, other):
        r"""Return the commutator product.

        See (1.1.55) in [HS].

        Often written :math:`A\times B`.
        """
        return (self*other - other*self)/2

    def __pow__(self, other):
        """Return *self* to the integer power *other*."""

        other = int(other)

        from pymbolic.algorithm import integer_power
        return integer_power(self, other, one=MultiVector({0: 1}, self.space))

    def __truediv__(self, other):
        """Return ``self*(1/other)``.
        """
        other = _cast_to_mv(other, self.space)

        return self*other.inv()

    def __rtruediv__(self, other):
        """Return ``other * (1/self)``.
        """
        other = MultiVector(other, self.space)

        return other * self.inv()

    __div__ = __truediv__

    # }}}

    # {{{ unary operations

    def inv(self):
        """Return the *multiplicative inverse* of the blade *self*.

        Often written :math:`A^{-1}`.
        """

        nsqr = self.norm_squared()
        if len(self.data) == 0:
            raise ZeroDivisionError
        if len(self.data) > 1:
            if self.get_pure_grade() in [0, 1, self.space.dimensions]:
                return MultiVector({
                    bits: coeff/nsqr for bits, coeff in self.data.items()},
                    self.space)

            else:
                raise NotImplementedError("division by non-blades")

        (bits, coeff), = self.data.items()

        # (1.1.54) in [HS]
        grade = bits.bit_count()
        if grade*(grade-1)//2 % 2:
            coeff = -coeff

        coeff = coeff/nsqr

        return MultiVector({bits: coeff}, self.space)

    def rev(self):
        r"""Return the *reverse* of *self*, i.e. the multivector obtained by
        reversing the order of all component blades.

        Often written :math:`A^\dagger`.
        """
        new_data = {}
        for bits, coeff in self.data.items():
            grade = bits.bit_count()
            if grade*(grade-1)//2 % 2 == 0:
                new_data[bits] = coeff
            else:
                new_data[bits] = -coeff

        return MultiVector(new_data, self.space)

    def invol(self):
        r"""Return the grade involution (see Section 2.9.5 of [DFM]), i.e.
        all odd-grade blades have their signs flipped.

        Often written :math:`\widehat A`.
        """
        new_data = {}
        for bits, coeff in self.data.items():
            grade = bits.bit_count()
            if grade % 2 == 0:
                new_data[bits] = coeff
            else:
                new_data[bits] = -coeff

        return MultiVector(new_data, self.space)

    def dual(self):
        r"""Return the dual of *self*, see (1.2.26) in [HS].

        Written :math:`\widetilde A` by [HS] and :math:`A^\ast` by [DFW].
        """

        return self | self.I.rev()

    def __inv__(self):
        """Return the dual of *self*, see :meth:`dual`."""
        return self.dual()

    def norm_squared(self):
        return self.rev().scalar_product(self)

    def __abs__(self):
        return self.norm_squared()**0.5

    @property
    def I(self):  # noqa
        """Return the pseudoscalar associated with this object's :class:`Space`.
        """
        return MultiVector({2**self.space.dimensions-1: 1}, self.space)

    # }}}

    # {{{ comparisons

    @memoize_method
    def __hash__(self):
        result = hash(self.space)
        for bits, coeff in self.data.items():
            result ^= hash(bits) ^ hash(coeff)

        return result

    def __bool__(self):
        return bool(self.data)

    __nonzero__ = __bool__

    def __eq__(self, other):
        other = _cast_to_mv(other, self.space)

        return self.data == other.data

    def __ne__(self, other):
        return not self.__eq__(other)

    def zap_near_zeros(self, tol=None):
        """Remove blades whose coefficient is close to zero
        relative to the norm of *self*.
        """

        if tol is None:
            tol = 1e-12

        new_data = {}
        for bits, coeff in self.data.items():
            if abs(coeff) > tol:
                new_data[bits] = coeff

        return MultiVector(new_data, self.space)

    def close_to(self, other, tol=None):
        return not (self-other).zap_near_zeros(tol=tol)

    # }}}

    # {{{ grade manipulation

    def gen_blades(self, grade=None):
        """Generate all blades in *self*, optionally only those of a specific
        *grade*.
        """

        if grade is None:
            for bits, coeff in self.data.items():
                yield MultiVector({bits: coeff}, self.space)
        else:
            for bits, coeff in self.data.items():
                if bits.bit_count() == grade:
                    yield MultiVector({bits: coeff}, self.space)

    def project(self, r):
        r"""Return a new multivector containing only the blades of grade *r*.

        Often written :math:`\langle A\rangle_r`.
        """
        new_data = {}
        for bits, coeff in self.data.items():
            if bits.bit_count() == r:
                new_data[bits] = coeff

        return MultiVector(new_data, self.space)

    def xproject(self, r, dtype=None):
        r"""If ``r == 0``, return ``self.project(0).as_scalar()``.
        If ``r == 1``, return ``self.project(1).as_vector(dtype)``.
        Otherwise, return ``self.project(r)``.
        """
        if r == 0:
            return self.project(0).as_scalar()
        elif r == 1:
            return self.project(1).as_vector(dtype)
        else:
            return self.project(r)

    def all_grades(self):
        """Return a :class:`set` of grades occurring in *self*."""

        return {bits.bit_count() for bits, coeff in self.data.items()}

    def get_pure_grade(self):
        """If *self* only has components of a single grade, return
        that as an integer. Otherwise, return *None*.
        """
        if not self.data:
            return 0

        result = None

        for bits in self.data.keys():
            grade = bits.bit_count()
            if result is None:
                result = grade
            elif result == grade:
                pass
            else:
                return None

        return result

    def odd(self):
        """Extract the odd-grade blades."""
        new_data = {}
        for bits, coeff in self.data.items():
            if bits.bit_count() % 2:
                new_data[bits] = coeff

        return MultiVector(new_data, self.space)

    def even(self):
        """Extract the even-grade blades."""
        new_data = {}
        for bits, coeff in self.data.items():
            if bits.bit_count() % 2 == 0:
                new_data[bits] = coeff

        return MultiVector(new_data, self.space)

    def project_min_grade(self):
        """
        .. versionadded:: 2014.2
        """

        return self.project(min(self.all_grades()))

    def project_max_grade(self):
        """
        .. versionadded:: 2014.2
        """

        return self.project(max(self.all_grades()))

    def as_scalar(self):
        result = 0
        for bits, coeff in self.data.items():
            if bits != 0:
                raise ValueError("multivector is not a scalar")
            result = coeff

        return result

    def as_vector(self, dtype=None):
        """Return a :mod:`numpy` vector corresponding to the grade-1
        :class:`MultiVector` *self*.

        If *self* is not grade-1, :exc:`ValueError` is raised.
        """

        if dtype is not None:
            result = np.zeros(self.space.dimensions, dtype=dtype)
        else:
            result = [0] * self.space.dimensions

        log_table = {2**i: i for i in range(self.space.dimensions)}
        try:
            for bits, coeff in self.data.items():
                result[log_table[bits]] = coeff
        except KeyError:
            raise ValueError("multivector is not a purely grade-1") from None

        if dtype is not None:
            return result
        else:
            return np.array(result)

    # }}}

    # {{{ helper functions

    def map(self, f: Callable[[CoeffT], CoeffT]) -> MultiVector[CoeffT]:
        """Return a new :class:`MultiVector` with coefficients mapped by
        function *f*, which takes a single coefficient as input and returns the
        new coefficient.
        """
        changed = False
        new_data = {}
        for bits, coeff in self.data.items():
            new_coeff = f(coeff)
            new_data[bits] = new_coeff
            if coeff is not new_coeff:
                changed = True

        if not changed:
            return self
        else:
            return MultiVector(new_data, self.space)

    # }}}

# }}}


def componentwise(f: Callable[[CoeffT], CoeffT], expr: T) -> T:
    """Apply function *f* componentwise to object arrays and
    :class:`MultiVector` instances. *expr* is also allowed to
    be a scalar.
    """

    if isinstance(expr, MultiVector):
        return cast(T, expr.map(f))

    from pytools.obj_array import obj_array_vectorize
    return obj_array_vectorize(f, expr)

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2022 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import pymbolic.interop.matchpy as m
import pymbolic.primitives as p


def test_replace_with_variadic_op():
    # Replace 'a * c' -> '6'

    from pytools import product

    w1 = p.StarWildcard("w1_star")
    a, b, c = p.variables("a b c")

    def a_times_c_is_6(w1_star):
        result_args = [6]
        for k, v in w1_star.items():
            result_args.extend([k]*v)

        return product(result_args)

    a_times_c_pattern = a * c * w1
    expr = a * b * c

    rule = m.make_replacement_rule(a_times_c_pattern, a_times_c_is_6)
    replaced_expr = m.replace_all(expr, [rule])
    assert replaced_expr == (6 * b)


def test_replace_with_non_commutative_op():
    # Replace 'f(a, b, c, ...)' -> 'g(d, ...)'
    w1 = p.StarWildcard("w1_star")
    a, b, c, d, f, g, x, y, z = p.variables("a b c d f g x y z")

    def replacer(w1_star):
        return g(d, *w1_star)

    rule = m.make_replacement_rule(f(a, b, c, w1), replacer)
    replaced_expr = m.replace_all(f(a, b, c, x, y, z), [rule])
    assert replaced_expr == g(d, x, y, z)


def test_replace_with_ternary_ops():
    from pymbolic import parse

    expr = parse("b < f(c, d)")
    rule = m.make_replacement_rule(
        p.Variable("f")(p.Variable("c"), p.DotWildcard("w1_")),
        lambda w1_: (w1_*42))
    assert m.replace_all(expr, [rule]) == parse("b < (42 * d)")

    expr = parse("b if f(c, d) else g(e)")
    rule = m.make_replacement_rule(
        p.If(parse("f(c, d)"),
             p.DotWildcard("w1_"),
             p.DotWildcard("w2_")),
        lambda w1_, w2_: p.If(parse("f(d, c)"), w2_, w1_))
    assert m.replace_all(expr, [rule]) == parse("g(e) if f(d, c) else b")


def test_make_subexpr_subst():
    from functools import reduce

    from pymbolic import parse
    from pymbolic.mapper.flattener import flatten

    subject = parse("a[k]*b[i, j]*c[i, j]*d[k]")
    pattern = parse("b[i, j]*c[i, j]") * p.StarWildcard("w1_")

    rule = m.make_replacement_rule(
        pattern,
        lambda w1_: (parse("subst(i, j)")
                     * (reduce(lambda acc, x: acc * (x[0] ** x[1]),
                               w1_.items(),
                               1)))
    )

    replaced_expr = m.replace_all(subject, [rule])

    ref_expr = flatten(parse("subst(i, j)*a[(k,)]*d[(k,)]"))
    assert flatten(replaced_expr) == ref_expr



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from functools import partial

import pymbolic.primitives as prim
from pymbolic.mapper.evaluator import EvaluationMapper


class SympyLikeMapperBase:

    def __call__(self, expr, *args, **kwargs):
        return self.rec(expr, *args, **kwargs)

    def rec(self, expr, *args, **kwargs):
        mro = list(type(expr).__mro__)
        dispatch_class = kwargs.pop("dispatch_class", type(self))

        while mro:
            method_name = "map_"+mro.pop(0).__name__

            try:
                method = getattr(dispatch_class, method_name)
            except AttributeError:
                pass
            else:
                return method(self, expr, *args, **kwargs)

        return self.not_supported(expr)

    def not_supported(self, expr):
        print(expr, expr.__class__.__mro__)
        raise NotImplementedError(
                "{} does not know how to map type '{}'".format(
                    type(self).__name__,
                    type(expr).__name__))


# {{{ sympy like -> pymbolic

class SympyLikeToPymbolicMapper(SympyLikeMapperBase):

    # {{{ utils

    def to_float(self, expr):
        return float(expr)

    def function_name(self, expr):
        # Given a symbolic function application f(x), return the name of f as a
        # string
        raise NotImplementedError()

    # }}}

    def map_Symbol(self, expr):  # noqa
        return prim.Variable(str(expr.name))

    def map_Rational(self, expr):  # noqa
        p, q = expr.p, expr.q

        num = self.rec(p)
        denom = self.rec(q)

        if prim.is_zero(denom-1):
            return num
        return prim.Quotient(num, denom)

    def map_Integer(self, expr):  # noqa
        return int(expr)

    def map_Add(self, expr):  # noqa
        return prim.Sum(tuple([self.rec(arg) for arg in expr.args]))

    def map_Mul(self, expr):  # noqa
        return prim.Product(tuple([self.rec(arg) for arg in expr.args]))

    def map_Pow(self, expr):  # noqa
        base, exp = expr.args
        return prim.Power(self.rec(base), self.rec(exp))

    def map_Subs(self, expr):  # noqa
        return prim.Substitution(self.rec(expr.expr),
                tuple([v.name for v in expr.variables]),
                tuple([self.rec(v) for v in expr.point]),
                )

    def map_Derivative(self, expr):  # noqa
        return prim.Derivative(self.rec(expr.expr),
                tuple([v.name for v in expr.variables]))

    def map_UnevaluatedExpr(self, expr):  # noqa
        return self.rec(expr.args[0])

    def not_supported(self, expr):
        if isinstance(expr, int):
            return expr
        elif getattr(expr, "is_Function", False):
            if self.function_name(expr) == "Indexed":
                args = [self.rec(arg) for arg in expr.args]
                if len(args) == 2:
                    return prim.Subscript(args[0], args[1])
                else:
                    return prim.Subscript(args[0], tuple(args[1:]))
            return prim.Variable(self.function_name(expr))(
                    *[self.rec(arg) for arg in expr.args])
        else:
            return SympyLikeMapperBase.not_supported(self, expr)

    def _comparison_operator(self, expr, operator=None):
        left = self.rec(expr.args[0])
        right = self.rec(expr.args[1])
        return prim.Comparison(left, operator, right)

    map_Equality = partial(_comparison_operator, operator="==")  # noqa: N815
    map_Unequality = partial(_comparison_operator, operator="!=")  # noqa: N815  # spellchecker: disable-line
    map_GreaterThan = partial(_comparison_operator, operator=">=")  # noqa: N815
    map_LessThan = partial(_comparison_operator, operator="<=")  # noqa: N815
    map_StrictGreaterThan = partial(_comparison_operator, operator=">")  # noqa: N815
    map_StrictLessThan = partial(_comparison_operator, operator="<")  # noqa: N815

# }}}


# {{{ pymbolic -> sympy like

class PymbolicToSympyLikeMapper(EvaluationMapper):

    @property
    def sym(self):
        raise NotImplementedError

    def raise_conversion_error(self, message):
        raise NotImplementedError

    def map_variable(self, expr):
        return self.sym.Symbol(expr.name)

    def map_constant(self, expr):
        return self.sym.sympify(expr)

    def map_floor_div(self, expr):
        return self.sym.floor(self.rec(expr.numerator) / self.rec(expr.denominator))

    def map_call(self, expr):
        if isinstance(expr.function, prim.Variable):
            func_name = expr.function.name
            try:
                func = getattr(self.sym.functions, func_name)
            except AttributeError:
                func = self.sym.Function(func_name)
            return func(*[self.rec(par) for par in expr.parameters])
        else:
            self.raise_conversion_error(expr)

    def map_subscript(self, expr):
        if isinstance(expr.aggregate, prim.Variable):
            return self.sym.Function("Indexed")(expr.aggregate.name,
                *(self.rec(idx) for idx in expr.index_tuple))
        else:
            self.raise_conversion_error(expr)

    def map_substitution(self, expr):
        return self.sym.Subs(self.rec(expr.child),
                tuple([self.sym.Symbol(v) for v in expr.variables]),
                tuple([self.rec(v) for v in expr.values]),
                )

    def map_if(self, expr):
        cond = self.rec(expr.condition)
        return self.sym.Piecewise((self.rec(expr.then), cond),
                                  (self.rec(expr.else_), True)
                                  )

    def map_comparison(self, expr):
        left = self.rec(expr.left)
        right = self.rec(expr.right)
        if expr.operator == "==":
            return self.sym.Equality(left, right)
        elif expr.operator == "!=":
            return self.sym.Unequality(left, right)  # spellchecker: disable-line
        elif expr.operator == "<":
            return self.sym.StrictLessThan(left, right)
        elif expr.operator == ">":
            return self.sym.StrictGreaterThan(left, right)
        elif expr.operator == "<=":
            return self.sym.LessThan(left, right)
        elif expr.operator == ">=":
            return self.sym.GreaterThan(left, right)
        else:
            raise NotImplementedError(f"Unknown operator '{expr.operator}'")

    def map_derivative(self, expr):
        return self.sym.Derivative(self.rec(expr.child),
                *[self.sym.Symbol(v) for v in expr.variables])

# }}}

# vim: fdm=marker



from __future__ import annotations


"""
Interoperability with :mod:`matchpy.functions` for pattern-matching and
term-rewriting.

.. autofunction:: match
.. autofunction:: match_anywhere
.. autofunction:: replace_all
.. autofunction:: make_replacement_rule


Internal API
^^^^^^^^^^^^

.. autoclass:: PymbolicOp
.. autoclass:: Wildcard
"""

__copyright__ = "Copyright (C) 2022 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import abc
from collections.abc import Callable, Iterable, Iterator, Mapping
from dataclasses import dataclass, field, fields
from functools import partial
from typing import ClassVar, Generic, TypeAlias, TypeVar

from matchpy import (
    Arity,
    Atom as BaseAtom,
    Expression,
    Operation,
    ReplacementRule,
    Wildcard as BaseWildcard,
)

import pymbolic.primitives as p
from pymbolic.typing import Scalar as PbScalar


ExprT: TypeAlias = Expression
ConstantT = TypeVar("ConstantT")
ToMatchpyT = Callable[[p.ExpressionNode], ExprT]
FromMatchpyT = Callable[[ExprT], p.ExpressionNode]


_NOT_OPERAND_METADATA = {"not_an_operand": True}

op_dataclass = dataclass(frozen=True, eq=True, repr=True)
non_operand_field = partial(field, metadata=_NOT_OPERAND_METADATA)


# {{{ Matchable expression types

@op_dataclass
class _Constant(BaseAtom, Generic[ConstantT]):
    value: ConstantT
    variable_name: str | None = None

    @property
    def head(self):
        return self

    def __lt__(self, other):
        # Used by matchpy internally to order subexpressions
        if not isinstance(other, Expression):
            return NotImplemented
        if type(other) is type(self):
            if self.value == other.value:
                return (self.variable_name or "") < (other.variable_name or "")
            return str(self.value) < str(other.value)
        return type(self).__name__ < type(other).__name__


@op_dataclass
class Scalar(_Constant[PbScalar]):
    _mapper_method: str = "map_scalar"


@op_dataclass
class Id(_Constant[str]):
    pass


@op_dataclass
class ComparisonOp(_Constant[str]):
    pass


@op_dataclass
class TupleOp(Operation):
    _operands: tuple[ExprT, ...]
    variable_name: str | None = non_operand_field(default=None)

    arity: ClassVar[Arity] = Arity.variadic
    name: ClassVar[str] = "tuple"
    _mapper_method: ClassVar[str] = "map_tuple_op"
    unpacked_args_to_init: ClassVar[bool] = True

    @property
    def operands(self):
        return self._operands


@op_dataclass
class PymbolicOp(abc.ABC, Operation):
    """
    A base class for all pymbolic-like operations.
    """
    unpacked_args_to_init: ClassVar[bool] = True

    @abc.abstractproperty
    def variable_name(self):
        pass

    @property
    def operands(self) -> tuple[Expression, ...]:
        return tuple(getattr(self, field.name)
                     for field in fields(self)
                     if not field.metadata.get("not_an_operand", False))

    def __lt__(self, other):
        if type(other) is type(self):
            if self.operands == other.operands:
                return (self.variable_name or "") < (other.variable_name or "")
            return str(self.operands) < str(other.operands)
        return type(self).__name__ < type(other).__name__

    @property
    def name(self) -> str:
        return self.__class__.__name__


@op_dataclass
class Variable(PymbolicOp):
    id: Id
    arity: ClassVar[Arity] = Arity.unary
    variable_name: str | None = non_operand_field(default=None)
    _mapper_method: ClassVar[str] = "map_variable"


@op_dataclass
class Call(PymbolicOp):
    function: ExprT
    args: TupleOp
    variable_name: str | None = non_operand_field(default=None)

    arity: ClassVar[Arity] = Arity.binary
    _mapper_method: ClassVar[str] = "map_call"


@op_dataclass
class Subscript(PymbolicOp):
    aggregate: ExprT
    indices: TupleOp
    variable_name: str | None = non_operand_field(default=None)

    arity: ClassVar[Arity] = Arity.binary
    _mapper_method: ClassVar[str] = "map_subscript"


# {{{ binary ops

@op_dataclass
class _BinaryOp(PymbolicOp):
    x1: ExprT
    x2: ExprT

    arity: ClassVar[Arity] = Arity.binary
    variable_name: str | None = non_operand_field(default=None)


@op_dataclass
class TrueDiv(_BinaryOp):
    _mapper_method: ClassVar[str] = "map_true_div"


@op_dataclass
class FloorDiv(_BinaryOp):
    _mapper_method: ClassVar[str] = "map_floor_div"


@op_dataclass
class Modulo(_BinaryOp):
    _mapper_method: ClassVar[str] = "map_modulo"


@op_dataclass
class Power(_BinaryOp):
    _mapper_method: ClassVar[str] = "map_power"


@op_dataclass
class LeftShift(_BinaryOp):
    _mapper_method: ClassVar[str] = "map_left_shift"


@op_dataclass
class RightShift(_BinaryOp):
    _mapper_method: ClassVar[str] = "map_right_shift"

# }}}


# {{{ variadic ops

variadic_op_dataclass = dataclass(init=False, frozen=True, repr=True)


@variadic_op_dataclass
class _VariadicCommAssocOp(PymbolicOp):
    children: tuple[ExprT, ...]
    variable_name: str | None = non_operand_field(default=None)

    commutative: ClassVar[bool] = True
    associative: ClassVar[bool] = True
    arity: ClassVar[Arity] = Arity.variadic

    def __init__(self, *children: ExprT, variable_name=None):
        object.__setattr__(self, "children", children)
        object.__setattr__(self, "variable_name", variable_name)

    @property
    def operands(self) -> tuple[ExprT, ...]:
        return self.children


@variadic_op_dataclass
class Sum(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_sum"


@variadic_op_dataclass
class Product(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_product"


@variadic_op_dataclass
class LogicalOr(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_logical_or"


@variadic_op_dataclass
class LogicalAnd(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_logical_and"


@variadic_op_dataclass
class BitwiseOr(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_bitwise_or"


@variadic_op_dataclass
class BitwiseAnd(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_bitwise_and"


@variadic_op_dataclass
class BitwiseXor(_VariadicCommAssocOp):
    _mapper_method: ClassVar[str] = "map_bitwise_xor"

# }}}


# {{{ unary op

@op_dataclass
class _UnaryOp(PymbolicOp):
    x: ExprT
    arity: ClassVar[Arity] = Arity.unary
    variable_name: str | None = non_operand_field(default=None)


@op_dataclass
class LogicalNot(_UnaryOp):
    _mapper_method: ClassVar[str] = "map_logical_not"


@op_dataclass
class BitwiseNot(_UnaryOp):
    _mapper_method: ClassVar[str] = "map_bitwise_not"

# }}}


@op_dataclass
class Comparison(PymbolicOp):
    left: ExprT
    operator: ComparisonOp
    right: ExprT
    variable_name: str | None = non_operand_field(default=None)

    arity: ClassVar[Arity] = Arity.ternary
    _mapper_method: ClassVar[str] = "map_comparison"


@op_dataclass
class If(PymbolicOp):
    condition: ExprT
    then: ExprT
    else_: ExprT
    variable_name: str | None = non_operand_field(default=None)

    arity: ClassVar[Arity] = Arity.ternary
    _mapper_method: ClassVar[str] = "map_if"


class Wildcard(BaseWildcard):

    # {{{ FIXME: This should go into matchpy itself.

    @classmethod
    def dot(cls, name=None) -> Wildcard:
        return cls(min_count=1, fixed_size=True, variable_name=name)

    @classmethod
    def star(cls, name=None) -> Wildcard:
        # FIXME: This should go into matchpy itself.
        return cls(min_count=0, fixed_size=False, variable_name=name)

    @classmethod
    def plus(cls, name=None) -> Wildcard:
        # FIXME: This should go into matchpy itself.
        return cls(min_count=1, fixed_size=False, variable_name=name)

    # }}}

# }}}


def _get_operand_at_path(expr: PymbolicOp, path: tuple[int, ...]) -> PymbolicOp:
    result = expr

    while path:
        step, path = path[0], path[1:]
        result = result.operands[step]

    return result


def match(subject: p.ExpressionNode,
          pattern: p.ExpressionNode,
          to_matchpy_expr: ToMatchpyT | None = None,
          from_matchpy_expr: FromMatchpyT | None = None
          ) -> Iterator[Mapping[str, p.ExpressionNode | PbScalar]]:
    from matchpy import Pattern, match

    from .tofrom import FromMatchpyExpressionMapper, ToMatchpyExpressionMapper

    if to_matchpy_expr is None:
        to_matchpy_expr = ToMatchpyExpressionMapper()
    if from_matchpy_expr is None:
        from_matchpy_expr = FromMatchpyExpressionMapper()

    m_subject = to_matchpy_expr(subject)
    m_pattern = Pattern(to_matchpy_expr(pattern))
    matches = match(m_subject, m_pattern)

    for subst in matches:
        yield {name: from_matchpy_expr(expr)
               for name, expr in subst.items()}


def match_anywhere(subject: p.ExpressionNode,
                   pattern: p.ExpressionNode,
                   to_matchpy_expr: ToMatchpyT | None = None,
                   from_matchpy_expr: FromMatchpyT | None = None
                   ) -> Iterator[tuple[Mapping[str, p.ExpressionNode | PbScalar],
                                       p.ExpressionNode | PbScalar]
                                 ]:
    from matchpy import Pattern, match_anywhere

    from .tofrom import FromMatchpyExpressionMapper, ToMatchpyExpressionMapper

    if to_matchpy_expr is None:
        to_matchpy_expr = ToMatchpyExpressionMapper()
    if from_matchpy_expr is None:
        from_matchpy_expr = FromMatchpyExpressionMapper()

    m_subject = to_matchpy_expr(subject)
    m_pattern = Pattern(to_matchpy_expr(pattern))
    matches = match_anywhere(m_subject, m_pattern)

    for subst, path in matches:
        yield ({name: from_matchpy_expr(expr)
                for name, expr in subst.items()},
               from_matchpy_expr(_get_operand_at_path(m_subject, path)))


def make_replacement_rule(pattern: p.ExpressionNode,
                          replacement: Callable[..., p.ExpressionNode],
                          to_matchpy_expr: ToMatchpyT | None = None,
                          from_matchpy_expr: FromMatchpyT | None = None
                          ) -> ReplacementRule:
    """
    Returns a :class:`matchpy.functions.ReplacementRule` from the objects
    declared via :mod:`pymbolic.primitives` instances.
    """
    from matchpy import Pattern

    from .tofrom import (
        FromMatchpyExpressionMapper,
        ToFromReplacement,
        ToMatchpyExpressionMapper,
    )

    if to_matchpy_expr is None:
        to_matchpy_expr = ToMatchpyExpressionMapper()
    if from_matchpy_expr is None:
        from_matchpy_expr = FromMatchpyExpressionMapper()

    m_pattern = Pattern(to_matchpy_expr(pattern))
    return ReplacementRule(m_pattern, ToFromReplacement(replacement,
                                                        to_matchpy_expr,
                                                        from_matchpy_expr))


def replace_all(expression: p.ExpressionNode,
                rules: Iterable[ReplacementRule],
                to_matchpy_expr: ToMatchpyT | None = None,
                from_matchpy_expr: FromMatchpyT | None = None
                ) -> p.ExpressionNode | tuple[p.ExpressionNode, ...]:
    import collections.abc as abc

    from matchpy import replace_all

    from .tofrom import FromMatchpyExpressionMapper, ToMatchpyExpressionMapper

    if to_matchpy_expr is None:
        to_matchpy_expr = ToMatchpyExpressionMapper()
    if from_matchpy_expr is None:
        from_matchpy_expr = FromMatchpyExpressionMapper()

    m_expr = to_matchpy_expr(expression)
    result = replace_all(m_expr, rules)
    if isinstance(result, abc.Sequence):
        return tuple(from_matchpy_expr(e) for e in result)
    else:
        return from_matchpy_expr(result)

# vim: fdm=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from functools import partial

from pytools import module_getattr_for_deprecations

from . import compiler, parser, primitives
from .compiler import compile
from .mapper import (
    dependency,
    differentiator,
    distributor,
    evaluator,
    flattener,
    stringifier,
    substitutor,
)
from .mapper.differentiator import differentiate, differentiate as diff
from .mapper.distributor import distribute, distribute as expand
from .mapper.evaluator import evaluate, evaluate_kw
from .mapper.flattener import flatten
from .mapper.substitutor import substitute
from .parser import parse
from .primitives import (  # noqa: N813
    ExpressionNode,
    Variable,
    Variable as var,
    disable_subscript_by_getitem,
    expr_dataclass,
    flattened_product,
    flattened_sum,
    linear_combination,
    make_common_subexpression as cse,
    make_sym_vector,
    quotient,
    subscript,
    variables,
)
from .typing import (
    ArithmeticExpression,
    Bool,
    Expression,
    Expression as _TypingExpression,
    Number,
    Scalar,
)
from pymbolic.version import VERSION_TEXT as __version__  # noqa


__all__ = (
    "ArithmeticExpression",
    "Bool",
    "Expression",
    "ExpressionNode",
    "Number",
    "Scalar",
    "Variable",
    "compile",
    "compiler",
    "cse",
    "dependency",
    "diff",
    "differentiate",
    "differentiator",
    "disable_subscript_by_getitem",
    "distribute",
    "distributor",
    "evaluate",
    "evaluate_kw",
    "evaluator",
    "expand",
    "expr_dataclass",
    "flatten",
    "flattened_product",
    "flattened_sum",
    "flattener",
    "linear_combination",
    "make_sym_vector",
    "parse",
    "parser",
    "primitives",
    "quotient",
    "stringifier",
    "subscript",
    "substitute",
    "substitutor",
    "var",
    "variables",
)

__getattr__ = partial(module_getattr_for_deprecations, __name__, {
        "ExpressionT": ("pymbolic.typing.Expression", _TypingExpression, 2026),
        "ArithmeticExpressionT": ("ArithmeticExpression", ArithmeticExpression, 2026),
        "BoolT": ("Bool", Bool, 2026),
        "ScalarT": ("Scalar", Scalar, 2026),
        })



from __future__ import annotations


__copyright__ = "Copyright (C) 2014 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# This is experimental, undocumented, and could go away any second.
# Consider yourself warned.

from collections.abc import Hashable
from typing import ClassVar

from pymbolic.primitives import ExpressionNode, Variable, expr_dataclass
from pymbolic.typing import Expression


class MultiVectorVariable(Variable):
    mapper_method = "map_multivector_variable"


# {{{ geometric calculus

class _GeometricCalculusExpression(ExpressionNode):
    def stringifier(self):
        from pymbolic.geometric_algebra.mapper import StringifyMapper
        return StringifyMapper


@expr_dataclass()
class NablaComponent(_GeometricCalculusExpression):
    ambient_axis: int
    nabla_id: Hashable


@expr_dataclass()
class Nabla(_GeometricCalculusExpression):
    nabla_id: Hashable


@expr_dataclass()
class DerivativeSource(_GeometricCalculusExpression):
    operand: Expression
    nabla_id: Hashable


class Derivative:
    """This mechanism cannot be used to take more than one derivative at a time.

    .. autoproperty:: nabla
    .. automethod:: __call__
    .. automethod:: dnabla
    .. automethod:: resolve
    """
    _next_id: ClassVar[list[int]] = [0]

    def __init__(self):
        self.my_id = f"id{self._next_id[0]}"
        self._next_id[0] += 1

    @property
    def nabla(self):
        return Nabla(self.my_id)

    def dnabla(self, ambient_dim):
        from pytools.obj_array import make_obj_array

        from pymbolic.geometric_algebra import MultiVector
        return MultiVector(make_obj_array(
            [NablaComponent(axis, self.my_id)
                for axis in range(ambient_dim)]))

    def __call__(self, operand):
        from pymbolic.geometric_algebra import MultiVector
        if isinstance(operand, MultiVector):
            return operand.map(
                    lambda coeff: DerivativeSource(coeff, self.my_id))
        else:
            return DerivativeSource(operand, self.my_id)

    @staticmethod
    def resolve(expr):
        # This method will need to be overridden by codes using this
        # infrastructure to use the appropriate subclass of DerivativeBinder.

        from pymbolic.geometric_algebra.mapper import DerivativeBinder
        return DerivativeBinder()(expr)

# }}}

# vim: foldmethod=marker



"""
.. autoclass:: EvaluationMapper
.. autoclass:: CachedEvaluationMapper
.. autoclass:: FloatEvaluationMapper
.. autoclass:: CachedFloatEvaluationMapper

.. autofunction:: evaluate
.. autofunction:: evaluate_kw
.. autofunction:: evaluate_to_float
"""
from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import operator as op
from collections.abc import Mapping
from functools import reduce
from typing import TYPE_CHECKING, cast

import pymbolic.primitives as p
from pymbolic.mapper import CachedMapper, CSECachingMapperMixin, Mapper, ResultT
from pymbolic.typing import Expression


if TYPE_CHECKING:
    import numpy as np

    from pymbolic.geometric_algebra import MultiVector


class UnknownVariableError(Exception):
    pass


class EvaluationMapper(Mapper[ResultT, []], CSECachingMapperMixin):
    """Example usage:

    .. doctest::

        >>> import pymbolic.primitives as p
        >>> x = p.Variable("x")

        >>> u = 5*x**2 - 3*x
        >>> print(u)
        5*x**2 + (-1)*3*x

        >>> from pymbolic.mapper.evaluator import EvaluationMapper as EM
        >>> EM(context={"x": 5})(u)
        110
    """

    context: Mapping[str, ResultT]

    def __init__(self, context: Mapping[str, ResultT] | None = None) -> None:
        """
        :arg context: a mapping from variable names to values
        """
        if context is None:
            context = {}

        self.context = context

    def map_constant(self, expr: object) -> ResultT:
        return cast(ResultT, expr)

    def map_variable(self, expr: p.Variable) -> ResultT:
        try:
            return self.context[expr.name]
        except KeyError:
            raise UnknownVariableError(expr.name) from None

    def map_call(self, expr: p.Call) -> ResultT:
        return self.rec(expr.function)(*[self.rec(par) for par in expr.parameters])  # type: ignore[operator]

    def map_call_with_kwargs(self, expr: p.CallWithKwargs) -> ResultT:
        args = [self.rec(par) for par in expr.parameters]
        kwargs = {
                k: self.rec(v)
                for k, v in expr.kw_parameters.items()}

        return self.rec(expr.function)(*args, **kwargs)  # type: ignore[operator]

    def map_subscript(self, expr: p.Subscript) -> ResultT:
        return self.rec(expr.aggregate)[self.rec(expr.index)]  # type: ignore[index]

    def map_lookup(self, expr: p.Lookup) -> ResultT:
        return getattr(self.rec(expr.aggregate), expr.name)

    def map_sum(self, expr: p.Sum) -> ResultT:
        return sum(self.rec(child) for child in expr.children)  # type: ignore[return-value, misc]

    def map_product(self, expr: p.Product) -> ResultT:
        from pytools import product
        return product(self.rec(child) for child in expr.children)

    def map_quotient(self, expr: p.Quotient) -> ResultT:
        return self.rec(expr.numerator) / self.rec(expr.denominator)  # type: ignore[operator]

    def map_floor_div(self, expr: p.FloorDiv) -> ResultT:
        return self.rec(expr.numerator) // self.rec(expr.denominator)  # type: ignore[operator]

    def map_remainder(self, expr: p.Remainder) -> ResultT:
        return self.rec(expr.numerator) % self.rec(expr.denominator)  # type: ignore[operator]

    def map_power(self, expr: p.Power) -> ResultT:
        return self.rec(expr.base) ** self.rec(expr.exponent)  # type: ignore[operator]

    def map_left_shift(self, expr: p.LeftShift) -> ResultT:
        return self.rec(expr.shiftee) << self.rec(expr.shift)  # type: ignore[operator]

    def map_right_shift(self, expr: p.RightShift) -> ResultT:
        return self.rec(expr.shiftee) >> self.rec(expr.shift)  # type: ignore[operator]

    def map_bitwise_not(self, expr: p.BitwiseNot) -> ResultT:
        # ??? Why, pylint, why ???
        # pylint: disable=invalid-unary-operand-type
        return ~self.rec(expr.child)  # type: ignore[operator]

    def map_bitwise_or(self, expr: p.BitwiseOr) -> ResultT:
        return reduce(op.or_, (self.rec(ch) for ch in expr.children))

    def map_bitwise_xor(self, expr: p.BitwiseXor) -> ResultT:
        return reduce(op.xor, (self.rec(ch) for ch in expr.children))

    def map_bitwise_and(self, expr: p.BitwiseAnd) -> ResultT:
        return reduce(op.and_, (self.rec(ch) for ch in expr.children))

    def map_logical_not(self, expr: p.LogicalNot) -> bool:  # type: ignore[override]
        return not self.rec(expr.child)

    def map_logical_or(self, expr: p.LogicalOr) -> bool:  # type: ignore[override]
        return any(self.rec(ch) for ch in expr.children)

    def map_logical_and(self, expr: p.LogicalAnd) -> bool:  # type: ignore[override]
        return all(self.rec(ch) for ch in expr.children)

    def map_list(self, expr: list[Expression]) -> ResultT:
        return [self.rec(child) for child in expr]  # type: ignore[return-value]

    def map_numpy_array(self, expr: np.ndarray) -> ResultT:
        import numpy
        result = numpy.empty(expr.shape, dtype=object)
        for i in numpy.ndindex(expr.shape):
            result[i] = self.rec(expr[i])
        return result  # type: ignore[return-value]

    def map_multivector(self, expr: MultiVector) -> ResultT:
        return expr.map(lambda ch: self.rec(ch))  # type: ignore[return-value]

    def map_common_subexpression_uncached(self, expr: p.CommonSubexpression) -> ResultT:
        return self.rec(expr.child)

    def map_if(self, expr: p.If) -> ResultT:
        if self.rec(expr.condition):
            return self.rec(expr.then)
        else:
            return self.rec(expr.else_)

    def map_comparison(self, expr: p.Comparison) -> ResultT:
        import operator
        return getattr(operator, expr.operator_to_name[expr.operator])(
            self.rec(expr.left), self.rec(expr.right))

    def map_min(self, expr: p.Min) -> ResultT:
        return min(self.rec(child) for child in expr.children)  # type: ignore[type-var]

    def map_max(self, expr: p.Max) -> ResultT:
        return max(self.rec(child) for child in expr.children)  # type: ignore[type-var]

    def map_tuple(self, expr: tuple[Expression, ...]) -> ResultT:
        return tuple([self.rec(child) for child in expr])  # type: ignore[return-value]

    def map_nan(self, expr: p.NaN) -> ResultT:
        if expr.data_type is None:
            from math import nan
            return nan  # type:ignore[return-value]
        else:
            return expr.data_type(float("nan"))


class CachedEvaluationMapper(CachedMapper[ResultT, []], EvaluationMapper[ResultT]):
    def __init__(self, context=None):
        CachedMapper.__init__(self)
        EvaluationMapper.__init__(self, context=context)


class FloatEvaluationMapper(EvaluationMapper[float]):
    def map_constant(self, expr) -> float:
        return float(expr)

    def map_rational(self, expr) -> float:
        return self.rec(expr.numerator) / self.rec(expr.denominator)


class CachedFloatEvaluationMapper(CachedEvaluationMapper[float]):
    def map_constant(self, expr) -> float:
        return float(expr)

    def map_rational(self, expr) -> float:
        return self.rec(expr.numerator) / self.rec(expr.denominator)


def evaluate(
            expression: Expression,
            context: Mapping[str, ResultT] | None = None,
            mapper_cls: type[EvaluationMapper[ResultT]] = CachedEvaluationMapper,
        ) -> ResultT:
    """
    :arg mapper_cls: A :class:`type` of the evaluation mapper
        whose instance performs the evaluation.
    """
    if context is None:
        context = {}
    return mapper_cls(context)(expression)


def evaluate_kw(
            expression: Expression,
            mapper_cls: type[EvaluationMapper[ResultT]] = CachedEvaluationMapper,
            **context: ResultT,
        ) -> ResultT:
    """
    :arg mapper_cls: A :class:`type` of the evaluation mapper
        whose instance performs the evaluation.
    """
    return mapper_cls(context)(expression)


def evaluate_to_float(expression, context=None,
                      mapper_cls=CachedFloatEvaluationMapper) -> float:
    """
    :arg mapper_cls: A :class:`type` of the evaluation mapper
        whose instance performs the evaluation.
    """
    if context is None:
        context = {}
    return mapper_cls(context)(expression)



"""
.. autoclass:: DependencyMapper
.. autoclass:: CachedDependencyMapper
"""

from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from collections.abc import Set
from typing import Literal, TypeAlias

import pymbolic.primitives as p
from pymbolic.mapper import CachedMapper, Collector, CSECachingMapperMixin, P


DependenciesT: TypeAlias = Set[p.AlgebraicLeaf | p.CommonSubexpression]


class DependencyMapper(
    CSECachingMapperMixin[DependenciesT, P],
    Collector[p.AlgebraicLeaf | p.CommonSubexpression, P],
):
    """Maps an expression to the :class:`set` of expressions it
    is based on. The ``include_*`` arguments to the constructor
    determine which types of objects occur in this output set.
    If all are *False*, only :class:`pymbolic.primitives.Variable`
    instances are included.
    """

    def __init__(
        self,
        include_subscripts: bool = True,
        include_lookups: bool = True,
        include_calls: bool | Literal["descend_args"] = True,
        include_cses: bool = False,
        composite_leaves: bool | None = None,
    ) -> None:
        """
        :arg composite_leaves: Setting this is equivalent to setting
            all preceding ``include_*`` flags.
        """

        if composite_leaves is False:
            include_subscripts = False
            include_lookups = False
            include_calls = False

        if composite_leaves is True:
            include_subscripts = True
            include_lookups = True
            include_calls = True

        assert include_calls in [True, False, "descend_args"]

        self.include_subscripts = include_subscripts
        self.include_lookups = include_lookups
        self.include_calls = include_calls
        self.include_cses = include_cses

    def map_variable(
        self, expr: p.Variable, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        return {expr}

    def map_call(
        self, expr: p.Call, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        if self.include_calls == "descend_args":
            return self.combine([
                self.rec(child, *args, **kwargs) for child in expr.parameters
            ])
        elif self.include_calls:
            return {expr}
        else:
            return super().map_call(expr, *args, **kwargs)

    def map_call_with_kwargs(
        self, expr: p.CallWithKwargs, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        if self.include_calls == "descend_args":
            return self.combine(
                [self.rec(child, *args, **kwargs) for child in expr.parameters]
                + [
                    self.rec(val, *args, **kwargs)
                    for name, val in expr.kw_parameters.items()
                ]
            )
        elif self.include_calls:
            return {expr}
        else:
            return super().map_call_with_kwargs(expr, *args, **kwargs)

    def map_lookup(
        self, expr: p.Lookup, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        if self.include_lookups:
            return {expr}
        else:
            return super().map_lookup(expr, *args, **kwargs)

    def map_subscript(
        self, expr: p.Subscript, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        if self.include_subscripts:
            return {expr}
        else:
            return super().map_subscript(expr, *args, **kwargs)

    def map_common_subexpression_uncached(
        self, expr: p.CommonSubexpression, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        if self.include_cses:
            return {expr}
        else:
            # FIXME: These look like mypy bugs, revisit
            return Collector.map_common_subexpression(self, expr, *args, **kwargs)  # type: ignore[return-value, arg-type]

    def map_slice(
        self, expr: p.Slice, *args: P.args, **kwargs: P.kwargs
    ) -> DependenciesT:
        return self.combine([
            self.rec(child, *args, **kwargs)
            for child in expr.children
            if child is not None
        ])

    def map_nan(self, expr: p.NaN, *args: P.args, **kwargs: P.kwargs) -> DependenciesT:
        return set()


class CachedDependencyMapper(CachedMapper[DependenciesT, P],
                             DependencyMapper[P]):
    def __init__(
        self,
        include_subscripts: bool = True,
        include_lookups: bool = True,
        include_calls: bool | Literal["descend_args"] = True,
        include_cses: bool = False,
        composite_leaves: bool | None = None,
    ) -> None:
        CachedMapper.__init__(self)
        DependencyMapper.__init__(
            self,
            include_subscripts=include_subscripts,
            include_lookups=include_lookups,
            include_calls=include_calls,
            include_cses=include_cses,
            composite_leaves=composite_leaves,
        )



from __future__ import annotations


__copyright__ = "Copyright (C) 2022 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from dataclasses import dataclass, replace

import numpy as np

from pytools import UniqueNameGenerator

import pymbolic.primitives as prim
from pymbolic.mapper import CachedIdentityMapper, IdentityMapper
from pymbolic.mapper.optimize import optimize_mapper


@dataclass(frozen=True, eq=True)
class RandomExpressionGeneratorContext:
    rng: np.random.Generator
    vng: UniqueNameGenerator
    current_depth: int
    max_depth: int

    def with_increased_depth(self):
        return replace(self, current_depth=self.current_depth+1)


def _generate_random_expr_inner(
        context: RandomExpressionGeneratorContext) -> prim.ExpressionNode:

    if context.current_depth >= context.max_depth:
        # force expression to be a leaf type
        return context.rng.integers(0, 42)

    bucket = context.rng.integers(0, 100) / 100.0

    # {{{ set some distribution of expression types

    # 'weight' is proportional to the probability of seeing an expression type
    weights = [1, 1, 1, 1, 1]
    expr_types = [prim.Variable, prim.Sum, prim.Product, prim.Quotient,
                  prim.Call]
    assert len(weights) == len(expr_types)

    # }}}

    buckets = np.cumsum(weights, dtype="float64")/np.sum(weights)

    expr_type = expr_types[np.searchsorted(buckets, bucket)]

    if expr_type == prim.Variable:
        return prim.Variable(context.vng("x"))
    elif expr_type in [prim.Sum, prim.Product]:
        left = _generate_random_expr_inner(context.with_increased_depth())
        right = _generate_random_expr_inner(context.with_increased_depth())
        return expr_type((left, right))
    elif expr_type == prim.Quotient:
        num = _generate_random_expr_inner(context.with_increased_depth())
        den = _generate_random_expr_inner(context.with_increased_depth())
        return prim.Quotient(num, den)
    elif expr_type == prim.Quotient:
        num = _generate_random_expr_inner(context.with_increased_depth())
        den = _generate_random_expr_inner(context.with_increased_depth())
        return prim.Quotient(num, den)
    elif expr_type == prim.Call:
        nargs = 3
        return prim.Variable(context.vng("f"))(
            *[_generate_random_expr_inner(context.with_increased_depth())
              for _ in range(nargs)])
    else:
        raise NotImplementedError(expr_type)


def generate_random_expression(seed: int, max_depth: int = 8) -> prim.ExpressionNode:
    from numpy.random import default_rng
    rng = default_rng(seed)
    vng = UniqueNameGenerator()

    context = RandomExpressionGeneratorContext(rng,
                                               vng=vng,
                                               max_depth=max_depth,
                                               current_depth=0)

    return _generate_random_expr_inner(context)


# {{{ custom mappers for tests

class _AlwaysFlatteningMixin:
    def map_sum(self, expr, *args, **kwargs):
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        from pymbolic.primitives import flattened_sum
        return flattened_sum(children)

    def map_product(self, expr, *args, **kwargs):
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        from pymbolic.primitives import flattened_product
        return flattened_product(children)


class AlwaysFlatteningIdentityMapper(_AlwaysFlatteningMixin,
                                     IdentityMapper):
    pass


class AlwaysFlatteningCachedIdentityMapper(_AlwaysFlatteningMixin,
                                           CachedIdentityMapper):
    pass

# }}}


# {{{ supporting bits for mapper optimizer test

BIG_EXPR_STR = """
(-1)*((cse_577[_pt_data_48[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0], _pt_data_49[(iface_ensm15*1075540 +
iel_ensm15*10 + idof_ensm15) % 10]] if _pt_data_48[((iface_ensm15*1075540 +
iel_ensm15*10 + idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_577[_pt_data_46[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_47[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_46[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_577[_pt_data_7[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_43[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_7[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_577[_pt_data_44[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_45[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_44[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_579[_pt_data_68[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_69[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_68[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_579[_pt_data_66[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_67[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_66[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_579[_pt_data_50[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_63[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_50[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_579[_pt_data_64[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_65[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_64[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_581[_pt_data_88[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_89[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_88[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_581[_pt_data_86[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_87[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_86[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_581[_pt_data_70[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_83[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_70[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_581[_pt_data_84[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_85[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_84[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_582[_pt_data_107[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_108[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_107[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_582[_pt_data_105[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_106[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_105[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_582[_pt_data_90[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_102[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_90[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_582[_pt_data_103[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_104[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_103[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0)) +
(cse_572[_pt_data_48[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_49[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_48[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_572[_pt_data_46[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_47[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_46[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_572[_pt_data_7[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_43[(iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 10]] if _pt_data_7[((iface_ensm15*1075540 + iel_ensm15*10 +
idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
(cse_572[_pt_data_44[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) %
4302160) // 10, 0], _pt_data_45[(iface_ensm15*1075540 + iel_ensm15*10 +
                                 idof_ensm15) % 10]] if
_pt_data_44[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15) % 4302160) //
            10, 0] != -1 else 0) + (cse_573[_pt_data_68[((iface_ensm15*1075540
                                                          + iel_ensm15*10 +
                                                          idof_ensm15) %
                                                         4302160) // 10, 0],
                                            _pt_data_69[(iface_ensm15*1075540 +
                                                         iel_ensm15*10 +
                                                         idof_ensm15) % 10]] if
                                    _pt_data_68[((iface_ensm15*1075540 +
                                                  iel_ensm15*10 + idof_ensm15)
                                                 % 4302160) // 10, 0] != -1
                                    else 0) +
            (cse_573[_pt_data_66[((iface_ensm15*1075540 + iel_ensm15*10 +
                                   idof_ensm15) % 4302160) // 10, 0],
                     _pt_data_67[(iface_ensm15*1075540 + iel_ensm15*10 +
                                  idof_ensm15) % 10]] if
             _pt_data_66[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15)
                          % 4302160) // 10, 0] != -1 else 0) +
             (cse_573[_pt_data_50[((iface_ensm15*1075540 + iel_ensm15*10 +
                                    idof_ensm15) % 4302160) // 10, 0],
                      _pt_data_63[(iface_ensm15*1075540 + iel_ensm15*10 +
                                   idof_ensm15) % 10]] if
              _pt_data_50[((iface_ensm15*1075540 + iel_ensm15*10 + idof_ensm15)
                           % 4302160) // 10, 0] != -1 else 0) +
              (cse_573[_pt_data_64[((iface_ensm15*1075540 + iel_ensm15*10 +
                                     idof_ensm15) % 4302160) // 10, 0],
                       _pt_data_65[(iface_ensm15*1075540 + iel_ensm15*10 +
                                    idof_ensm15) % 10]] if
               _pt_data_64[((iface_ensm15*1075540 + iel_ensm15*10 +
                             idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
               (cse_574[_pt_data_88[((iface_ensm15*1075540 + iel_ensm15*10 +
                                      idof_ensm15) % 4302160) // 10, 0],
                        _pt_data_89[(iface_ensm15*1075540 + iel_ensm15*10 +
                                     idof_ensm15) % 10]] if
                _pt_data_88[((iface_ensm15*1075540 + iel_ensm15*10 +
                              idof_ensm15) % 4302160) // 10, 0] != -1 else 0) +
                (cse_574[_pt_data_86[((iface_ensm15*1075540 + iel_ensm15*10 +
                                       idof_ensm15) % 4302160) // 10, 0],
                         _pt_data_87[(iface_ensm15*1075540 + iel_ensm15*10 +
                                      idof_ensm15) % 10]] if
                 _pt_data_86[((iface_ensm15*1075540 + iel_ensm15*10 +
                               idof_ensm15) % 4302160) // 10, 0] != -1 else 0)
                 + (cse_574[_pt_data_70[((iface_ensm15*1075540 + iel_ensm15*10
                                          + idof_ensm15) % 4302160) // 10, 0],
                            _pt_data_83[(iface_ensm15*1075540 + iel_ensm15*10 +
                                         idof_ensm15) % 10]] if
                    _pt_data_70[((iface_ensm15*1075540 + iel_ensm15*10 +
                                  idof_ensm15) % 4302160) // 10, 0] != -1 else
                    0) + (cse_574[_pt_data_84[((iface_ensm15*1075540 +
                                                iel_ensm15*10 + idof_ensm15) %
                                               4302160) // 10, 0],
                                  _pt_data_85[(iface_ensm15*1075540 +
                                               iel_ensm15*10 + idof_ensm15) %
                                              10]] if
                          _pt_data_84[((iface_ensm15*1075540 + iel_ensm15*10 +
                                        idof_ensm15) % 4302160) // 10, 0] != -1
                          else 0) +
                    (cse_575[_pt_data_107[((iface_ensm15*1075540 +
                                            iel_ensm15*10 + idof_ensm15) %
                                           4302160) // 10, 0],
                             _pt_data_108[(iface_ensm15*1075540 + iel_ensm15*10
                                           + idof_ensm15) % 10]] if
                     _pt_data_107[((iface_ensm15*1075540 + iel_ensm15*10 +
                                    idof_ensm15) % 4302160) // 10, 0] != -1
                     else 0) + (cse_575[_pt_data_105[((iface_ensm15*1075540 +
                                                       iel_ensm15*10 +
                                                       idof_ensm15) % 4302160)
                                                     // 10, 0],
                                        _pt_data_106[(iface_ensm15*1075540 +
                                                      iel_ensm15*10 +
                                                      idof_ensm15) % 10]] if
                                _pt_data_105[((iface_ensm15*1075540 +
                                               iel_ensm15*10 + idof_ensm15) %
                                              4302160) // 10, 0] != -1 else 0)
                                + (cse_575[_pt_data_90[((iface_ensm15*1075540 +
                                                         iel_ensm15*10 +
                                                         idof_ensm15) %
                                                        4302160) // 10, 0],
                                           _pt_data_102[(iface_ensm15*1075540 +
                                                         iel_ensm15*10 +
                                                         idof_ensm15) % 10]] if
                                   _pt_data_90[((iface_ensm15*1075540 +
                                                 iel_ensm15*10 + idof_ensm15) %
                                                4302160) // 10, 0] != -1 else
                                   0) +
                                (cse_575[_pt_data_103[((iface_ensm15*1075540 +
                                                        iel_ensm15*10 +
                                                        idof_ensm15) % 4302160)
                                                      // 10, 0],
                                         _pt_data_104[(iface_ensm15*1075540 +
                                                       iel_ensm15*10 +
                                                       idof_ensm15) % 10]] if
                                 _pt_data_103[((iface_ensm15*1075540 +
                                                iel_ensm15*10 + idof_ensm15) %
                                               4302160) // 10, 0] != -1 else 0)
"""


_replacements = {
        "iface_ensm15": prim.Variable("_0"),
        "iel_ensm15": prim.Variable("_1"),
        "idof_ensm15": prim.Variable("_2"),
        }


class Renamer(IdentityMapper):
    def map_variable(self, expr):
        return _replacements.get(expr.name, expr)


@optimize_mapper(drop_args=True, drop_kwargs=True,
                 inline_cache=True, inline_rec=True,
                 inline_get_cache_key=True)
class OptimizedRenamer(CachedIdentityMapper):
    def map_variable(self, expr):
        return _replacements.get(expr.name, expr)

    def get_cache_key(self, expr):
        # Must add 'type(expr)', to differentiate between python scalar types.
        # In Python, the following conditions are true: "hash(4) == hash(4.0)"
        # and "4 == 4.0", but their traversal results cannot be re-used.
        return (type(expr), expr)

# }}}


# vim: foldmethod=marker
